//
//   DVector class definition.  Copy on write version.
//
#ifndef DVECTOR_HH
#define DVECTOR_HH

#include <iosfwd>
#include "Complex.hh"

#ifdef __GNUG__
#pragma interface
#endif

/**  The data vector holds a variable length data vector of arbitrary 
  *  type. Methods are provided to convert data to a requested type 
  *  independent of the internal representation of the data.
  *  DVector are primarily used for time and frequency series.
  *  @memo Data Vector class.
  *  @author John G. Zweizig
  *  @version 2.1; Last modified: May 19, 2004
  */
class DVector {
public:

  /**  DVType enumerates the data types that may be held by a DVector.
    *  Some of the data vector types may not be implemented. 
    *  @memo DVector data type enumeration.
    */
  enum DVType {
    t_short,
    t_int,
    t_long,
    t_float,
    t_double,
    t_complex,
    t_dcomplex
  };

  /**  Data type used internally for size and index variables.
    *  @memo Size and index data type.
    */
  typedef unsigned long size_type;

public:
  /**  Initialize the base class element, $i.e.$ the data length.
    *  @memo Default constructor.
    */
  DVector(void) {}

  /**  Null base class destructor.
    *  @memo Destructor.
    */
  virtual ~DVector() {}

  //-------------------------------------  Accessors

  /**    An instance of a class derived from DVector is duplicated and 
    *          a pointer to the duplicate is returned. All data are copied.
    *  @memo   Duplicate a DVector.
    *  @return a pointer to the DVector clone.
    */
  virtual DVector* clone(void) const {return Extract(0, getLength());}
  virtual DVector* DupVector(void) const {return clone();}

  /**  Erase a substring from the vector.
    *  @memo   Erase a subset of a vector.
    *  @param  inx    Index of first element to be extracted.
    *  @param  length Number of data elements to be extracted.
    */
  virtual void Erase(size_type inx, size_type length) = 0;

  /**    A substring starting at the first element is extracted from the 
   *           DVector. The substring length is forced to be valid.
    *  @memo   Extract a subset of a vector.
    *  @param  length number of data elements to be extracted.
    *  @return a pointer to a new DVector containing the extracted data.
    */
  virtual DVector* Extract(size_type length) const {return Extract(0, length);}

  /**    A substring is extracted from the DVector. The substring
    *          length is forced to be valid.
    *  @memo   Extract a subset of a vector.
    *  @param  inx    index of first element to be extracted
    *  @param  length number of data elements to be extracted.
    *  @return a pointer to a new DVector containing the extracted data.
    */
  virtual DVector* Extract(size_type inx, size_type length) const=0;

  /**  A substring is extracted and decimated from the DVector. The 
    *  extracted vector contains the elements defined by #DVector[N*inc]#
    *  where N is all the integers #0 <= N < length#.
    *  @memo   Extract a subset of a vector.
    *  @param  inx    Starting index of the substring to be extracted
    *  @param  length Number of elements to be extracted
    *  @param  inc    Index increment between successive extracted elements.
    *  @return pointer to the extracted data vector.
    */
  virtual DVector* Extract(size_type inx, size_type length, size_type inc) const=0;

  /**  Return the data vector type as a DVector::DVType code.
    *  @memo   Get the data type
    *  @return a data vector type code.
    */ 
  virtual DVType  getType(void) const=0;

  /**    Get the data type name as a constant character string.
    *  @memo   Get the data type name.
    *  @return pointer to a constant character array containing the data 
    *          type name.
    */
  const char* getTypeName(void) const;

  /**    Test if the vector data type is t_short.
    *  @memo   Test for short int data.
    *  @return true if the vector data contains short integers.
    */
  bool S_data(void) const {return getType() == t_short;}

  /**  Test if the vector data type is t_float.
    *  @memo Test for float data.
    *  @return true if the vector data contains floats.
    */
  bool I_data(void) const {return getType() == t_int;}

  /**  Test if the vector data type is t_float.
    *  @memo Test for float data.
    *  @return true if the vector data contains floats.
    */
  bool F_data(void) const {return getType() == t_float;}

  /**  Test if the vector data type is t_float.
    *  @memo Test for float data.
    *  @return true if the vector data contains floats.
    */
  bool D_data(void) const {return getType() == t_double;}

  /**    Test if the vector data type is t_complex.
    *  @memo   Test for fComplex data.
    *  @return true if the vector data type is t_complex.
    */
  bool C_data(void) const {return getType() == t_complex;}

  /**  Get the number of bytes in a data word.
    *  @memo Get the size of a data word.
    *  @return The number of bytes in a data word.
    */
  virtual size_type getSize(void) const=0;

  /**  Get the number of data elements in the vector.
    *  @memo Get the data length.
    *  @return The data vector length.
    */
  virtual size_type getLength(void) const = 0;
  size_type size(void) const { return getLength();}
 

  /**  Return the capacity of the data vector in data words.
    *  @memo Get the vector capacity
    *  @return Number of words that can fit in the vector.
    */
  virtual size_type capacity(void) const = 0;

  /**  #Resize# insures that there is sufficient storage allocated 
    *  for #len# data items. No memory is release if the current 
    *  allocation is greater than the requested size. The vector
    *  data are copied if the vector storage must be reallocated.
    *  The current data length is set to the specified length on return
    *  but any new data remain uninitialized.
    *  @memo Expand the vector storage.
    *  @param len Minimum number of words that must be available,
    */
  virtual void ReSize(size_type len) = 0;

  /**  A data vector of the specified length is allocated. The current
    *  data length remains the same. If the present data vector is already
    *  greater than or equal to the specified length the data vector is not
    *  reallocated. If reallocation occurs, the current data may be copied
    *  to the front of the new vector. The vector is not reserved for write 
    *  access.
    *  @memo  Insure a minimum number of words.
    *  @param len Number of words to reserve.
    */
  virtual void reserve(size_type len) = 0;

  /**  Get a pointer to the data array. Note that because the DVector uses
    *  internal copy on write data storage, the data can be moved by any 
    *  manipulation of the data.
    *  @memo Get a pointer to the data.
    *  @return void pointer to the data array.
    */
  virtual void* refData(void) {return 0;}

  /**  Get a constant pointer to the data array. Note that because the DVector 
    *  uses internal copy-on-write data storage, the data can be moved by any 
    *  manipulation of the data.
    *  @memo Get a constant pointer to the data.
    *  @return constant void pointer to the data array.
    */
  virtual const void* refData(void) const {return 0;}

  /**  Copy data into a short integer output array. If the DVector has
    *        a different type than the output array, the data are 
    *        converted to the output type.
    *  @memo Copy/Convert data into a short integer array.
    *  @param inx   Starting index of data to be copied.
    *  @param Ndata Maximum number of data words to be copied.
    *  @param Data  Output short integer array.
    *  @return number of data words actually copied.
    */
  virtual size_type getData(size_type inx, size_type Ndata, short Data[]) const=0;

  /**  Copy data into an integer output array. If the DVector has
    *        a different type than the output array, the data are 
    *        converted to the output type.
    *  @memo Copy/Convert data into a float array.
    *  @param inx   Starting index of data to be copied.
    *  @param Ndata Maximum number of data words to be copied.
    *  @param Data  Output integer array.
    *  @return number of data words actually copied.
    */
  virtual size_type getData(size_type inx, size_type Ndata, int Data[]) const=0;

  /**  Copy data into a float output array. If the DVector has
    *        a different type than the output array, the data are 
    *        converted to the output type.
    *  @memo Copy/Convert data into a float array.
    *  @param inx   Starting index of data to be copied.
    *  @param Ndata Maximum number of data words to be copied.
    *  @param Data  Output float array.
    *  @return number of data words actually copied.
    */
  virtual size_type getData(size_type inx, size_type Ndata, float Data[]) const=0;

  /**  Copy data into a double float output array. If the DVector has
    *        a different type than the output array, the data are 
    *        converted to the output type.
    *  @memo Copy/Convert data into a float array.
    *  @param inx   Starting index of data to be copied.
    *  @param Ndata Maximum number of data words to be copied.
    *  @param Data  Output double float array.
    *  @return number of data words actually copied.
    */
  virtual size_type getData(size_type inx, size_type Ndata, double Data[]) const=0;

  /**  Copy data into a complex output array. If the DVector has
    *        a different type than the output array, the data are 
    *        converted to the output type.
    *  @memo Copy/Convert data into a complex array.
    *  @param inx   Starting index of data to be copied.
    *  @param Ndata Maximum number of data words to be copied.
    *  @param Data  Output complex array.
    *  @return number of data words actually copied.
    */
  virtual size_type getData(size_type inx, size_type Ndata, fComplex Data[]) const=0;

  /**    A specified data word is picked up from the DVector, converted 
    *          to a short integer and returned. No validity checking is made 
    *          on the index.
    *  @memo   Get one data word.
    *  @param  inx Index into the data vector of the word to be returned.
    *  @return value of selected word converted if necessary to short integer.
    */
  virtual short getShort(size_type inx) const=0;

  /**    A specified data word is picked up from the DVector, converted 
    *          to an integer and returned. No validity checking is made 
    *          on the index.
    *  @memo   Get one data word.
    *  @param  inx Index into the data vector of the word to be returned.
    *  @return value of selected word converted if necessary to integer.
    */
  virtual int getInt(size_type inx) const=0;

  /**    A specified data word is picked up from the DVector, converted 
    *          to a float and returned. No validity checking is made 
    *          on the index.
    *  @memo   Get one data word.
    *  @param  inx Index into the data vector of the word to be returned.
    *  @return value of selected work converted if necessary to float.
    */
  virtual float getFloat(size_type inx) const=0;

  /**    A specified data word is picked up from the DVector, converted 
    *          to a double float and returned. No validity checking is made 
    *          on the index.
    *  @memo   Get one data word.
    *  @param  inx Index into the data vector of the word to be returned.
    *  @return value of selected word converted if necessary to double.
    */
  virtual double getDouble(size_type inx) const=0;

  /**    A specified data word is picked up from the DVector, converted 
    *          to a complex and returned. No validity checking is made 
    *          on the index.
    *  @memo   Get one data word.
    *  @param  inx Index into the data vector of the word to be returned.
    *  @return value of selected work converted if necessary to complex.
    */
  virtual fComplex getCplx(size_type inx)  const=0;

  /**  Return the largest entry. If the vector is complex, the largest 
    *  real part is returned.
    *  @memo   Maximum data value.
    *  @return Largest data value.
    */
  virtual double getMaximum(void)  const=0;

  /**  Return the smallest entry. If the vector is complex, the smallest 
    *  real part is returned.
    *  @memo   Minimum data value.
    *  @return Smallest data value.
    */
  virtual double getMinimum(void) const=0;

  /**  Count the number of entries between two specified limits. If the 
    *  vector is complex, the real part is compared.
    *  @memo   Number of entries greater than limit.
    *  @param  limit Lower limit of entries to be counted.
    *  @return Number of entries greater than limit.
    */
  virtual size_type getNBetween(double low, double high) const=0;

  /**  Count the number of entries with value greater than the specified 
    *          limit. If the vector is complex, the real part is compared.
    *  @memo   Number of entries greater than limit.
    *  @param  limit Lower limit of entries to be counted.
    *  @return Number of entries greater than limit.
    */
  virtual size_type getNGreater(double limit) const=0;

  /**  Count the number of entries with value less than the specified 
    *          limit. If the vector is complex, the real part is compared.
    *  @memo   Number of entries less than limit.
    *  @param  limit Upper limit of entries to be counted.
    *  @return Number of entries less than limit.
    */
  virtual size_type getNLess(double limit) const=0;

  /**  Insert #N# values starting at #inx2# of the argument vector into the 
    *  current vector starting at index #inx#.
    *  @memo Inset data from a vector into the curret vector.
    *  @param inx  Index of the insertion point of the current vector.
    *  @param v    Vector containing data to be inserted.
    *  @param inx2 index in v of the data to be inserted.
    *  @param N    Number of words to be inserted
    *  @return Reference to the result vector.
    */
  virtual DVector& insert(size_type inx, const DVector& v, 
			  size_type inx2=0, size_type N=0);

  /**  Replace #N# values starting at #inx# of the current vector 
    *  with #nV# values starting at index #inv# of the argument vector.
    *  Note that the sub-strings must be within the range of the
    *  vectors, $i.e.$ #0 <= inx+N < len# and  #0 <= inv+Nv < v.len#.
    *  @memo Replace data from a sub-vector into the curret vector.
    *  @param inx  Index of the insertion point of the current vector.
    *  @param N    Number of words to be replaced
    *  @param v    Vector containing data to be inserted.
    *  @param inv  index in v of the data to be inserted.
    *  @param Nv   Number of words to be inserted
    *  @return Reference to the result vector.
    */
  virtual DVector& replace(size_type inx, size_type N, const DVector& v, 
			   size_type inv, size_type Nv) = 0;

  /**  Replace #N# values starting at #inx# of the current vector 
    *  with #N2# values starting at index #inx2# of the argument vector.
    *  @memo Replace data from a vector into the curret vector.
    *  @param inx  Index of the insertion point of the current vector.
    *  @param N    Number of words to be replaced
    *  @param v    Vector containing data to be inserted.
    *  @return Reference to the result vector.
    */
  DVector& replace(size_type inx, size_type N, const DVector& v);

  /**  Add #N# values in the addend vector to the current vector starting at
    *  index #inx# of the current vector and #inx2# of the addend vector.
    *  @memo Add two vectors
    *  @param inx  Index of the first element in this vector to be added.
    *  @param v    Vector to be added
    *  @param inx2 index in the 
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual DVector& add(size_type inx, const DVector& v, 
		       size_type inx2=0, size_type N=0) = 0;

  /**  Add a constant to #N# values of the current vector starting at
    *  index #inx#.
    *  @memo Add a constant to a vector
    *  @param inx  Index of the first element in this vector to be added.
    *  @param x    Constant to be added
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual DVector& bias(size_type inx, double x, size_type N=0) = 0;

  /**  Subtract #N# values in the subtrahend vector from the current vector 
    *  starting at index #inx# of the current vector and #inx2# of the 
    *  subtrahend vector.
    *  @memo Subtract two vectors
    *  @param inx  Index of the first element in this vector to be added.
    *  @param v    Vector to be added
    *  @param inx2 index in the 
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual DVector& sub(size_type inx, const DVector& v, 
		       size_type inx2=0, size_type N=0) = 0;

  /**  Multiply #N# values in the current vector starting at index #inx# 
   *   by the values starting at #inx2# of the multiplier vector.
    *  @memo Multiply two vectors
    *  @param inx  Index of the first element in this vector to be added.
    *  @param v    Vector to be added
    *  @param inx2 index in the 
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual DVector& mpy(size_type inx, const DVector& v, 
		       size_type inx2=0, size_type N=0) = 0;

  /**  Multiply #N# values of the current vector starting at index #inx#
    *  by a constant.
    *  @memo Multiply a vector by a constant
    *  @param inx  Index of the first element in this vector to be added.
    *  @param x    Constant to be added
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual DVector& scale(size_type inx, double x, size_type N=0) = 0;

  /**  Divide #N# values starting at index #inx# of the current vector by
    *  the values starting at #inx2# of the divisor vector.
    *  @memo Divide two vectors
    *  @param inx  Index of the first element in this vector to be added.
    *  @param v    Vector to be added
    *  @param inx2 index in the 
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual DVector& div(size_type inx, const DVector& v, 
		       size_type inx2=0, size_type N=0) = 0;

  /**  Inner product between two substrings.
    *  @memo Inner product
    *  @param inx  Index of the first element in this vector to be added.
    *  @param v    Vector to be added
    *  @param inx2 index in the 
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual double dot(size_type inx, const DVector& v, 
		     size_type inx2=0, size_type N=0) const = 0;

  /**  Complex intter p roduct of two subvectors.
    *  @memo Comples inner product.
    *  @param inx  Index of the first element in this vector to be added.
    *  @param v    Vector to be added
    *  @param inx2 index in the 
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  virtual fComplex cdot(size_type inx, const DVector& v, 
			size_type inx2=0, size_type N=0) const = 0;

  /**  All words in a specified range are converted to double and summed.
    *  @memo   Sum N data words.
    *  @param  inx Index into the data vector of first word to be summed.
    *  @param  N   Number of words to be summed.
    *  @return Sum of specified data vector.
    */
  virtual double VSum(size_type inx=0, size_type N=0) const=0;

  /**  All words in a specified range are converted to complex and summed.
    *  @memo   Sum N data words.
    *  @param  inx Index into the data vector of first word to be summed.
    *  @param  N   Number of words to be summed.
    *  @return Sum of specified data vector.
    */
  virtual fComplex CSum(size_type inx=0, size_type N=0) const=0;

  /**  Calculate the inner (dot) product of two DVectors. 
    *        The inner product is defined as the sum of the products of
    *        the corresponding elements, i.e. 
    *        $ X \cdot Y = \Sigma X_i \cdot Y_i $.
    *        If the right hand side operand is complex, it is conjugated 
    *        before the multiplication.
    *  @memo Inner product.
    *  @param  rhs DVector to be multiplied by the other.
    *  @return The dot product of two DVectors.
    */
  virtual double operator*(const DVector& rhs) const;

  /**   The data from the rhs DVector is converted and copied to the 
    *         lhs DVector. The original data in the lhs DVector are lost.
    *  @memo  Assignment operator.
    *  @param rhs Constant DVector whose value is copied.
    *  @return reference to the modified DVector.
    */
  virtual DVector& operator =(const DVector& rhs) = 0; 

  /**  The data from the rhs DVector is converted and compared to the
    *  lhs DVector. The original data are untouched.
    *  @memo  Comparison operator.
    *  @param rhs Constant DVector whose value is copied.
    *  @return reference to the modified DVector.
    */
  virtual bool operator==(const DVector& rhs) const = 0;

  /**  The data in the rhs DVector are added to the data of the lhs 
    *        DVector on an element by element basis.
    *  @memo Add a DVector.
    *  @param rhs DVector whose data are added.
    *  @return reference to the modified DVector.
    */
  virtual DVector& operator+=(const DVector& rhs);

  /**  The data in the rhs DVector are subtracted from the data of 
    *        the lhs DVector on an element by element basis.
    *  @memo Subtract a DVector.
    *  @param rhs DVector whose data are subtracted.
    *  @return reference to the modified DVector.
    */
  virtual DVector& operator-=(const DVector& rhs);

  /**  The right hand side (rhs) constant is added to each element of 
   *   the left hand side DVector.
    *  @memo Bias a DVector.
    *  @param rhs Scalar constant to be added to each element.
    *  @return  reference to the resulting DVector.
    */
  virtual DVector& operator+=(double rhs);

  /**  The right hand side constant (rhs) is subtracted to each element 
   *   of the left hand side DVector.
    *  @memo Bias a DVector (negative).
    *  @param rhs Scalar constant to be subtracted from each element.
    *  @return  reference to the resulting DVector.
    */
  DVector& operator-=(double rhs);

  /**  Each element of the lhs DVector is multiplied by the rhs constant.
    *  @memo Scale a DVector.
    *  @param rhs Scalar constant each element is multiplied by.
    *  @return reference to the resulting DVector.
    */
  virtual DVector& operator*=(double rhs);

  /**  Each element of the lhs DVector is divided by the rhs constant.
    *  @memo Scale a DVector (divide).
    *  @param rhs Scalar constant each element is divided by.
    *  @return reference to the resulting DVector.
    */
  virtual DVector& operator/=(double rhs);

  /**  Each element of the lhs DVector is multiplied by the corresponding 
    *  element of rhs.
    *  @memo Element by element multiplication of DVector.
    *  @param rhs Multiplier DVector.
    *  @return Reference to the resulting DVector.
    */
  virtual DVector& operator*=(const DVector& rhs);

  /**  Each element of the lhs DVector is divided by the corresponding 
    *  element of rhs.
    *  @memo Divide by a DVector.
    *  @param rhs Denominator DVector.
    *  @return Reference to the resulting DVector.
    */
  virtual DVector& operator/=(const DVector& rhs);

  /**  A formatted dump of the DVector contents is written to a 
    *        specified stream.
    *  @memo Dump the vector contents.
    *  @param str Output stream that is to receive the formatted dump.
    *  @return Reference to the output stream.
    */
  virtual std::ostream& Dump(std::ostream& out) const = 0;

  //---------------------------------------  Mutators
  /**  Data from a short integer array are converted to the DVector 
    *       type and appended to the data vector.
    *  @memo  Append short integer  data to a vector.
    *  @param N    Number of data words to append to the DVector.
    *  @param data short integer data array.
    */
  void Append(size_type N, const short* data);

  /**  Data from an integer array are converted to the DVector 
    *         type and appended to the data vector.
    *  @memo  Append integer data to a vector.
    *  @param N    Number of data words to append to the DVector.
    *  @param data integer data to be appended to the DVector.
    */
  void Append(size_type N, const int* data);

  /**  Data from a float array are converted to the DVector 
    *         type and appended to the data vector.
    *  @memo  Append float data to a vector.
    *  @param N    Number of data words to append to the DVector.
    *  @param data float data to be appended to the DVector.
    */
  void Append(size_type N, const float* data);

  /**  Data from a double float array are converted to the DVector 
    *         type and appended to the data vector.
    *  @memo  Append double float data to a vector.
    *  @param N    Number of data words to append to the DVector.
    *  @param data double float data to be appended to the DVector.
    */
  void Append(size_type N, const double* data);

  /**  Data from a complex array are converted to the DVector 
    *         type and appended to the data vector.
    *  @memo  Append complex data to a vector.
    *  @param N    Number of data words to append to the DVector.
    *  @param data complex data to be appended to the DVector.
    */
  void Append(size_type N, const fComplex* data);

  /**    Data from the specified DVector are appended to the object 
    *          vector. The appended data will be converted to the object 
    *          type if necessary.
    *  @memo   Append a DVector to a DVector.
    *  @param  data DVector containing data to be appended.
    */
  void Append(const DVector& data);

  /**  The current data length is set to zero. No allocated memory is 
    *        released.
    *  @memo Clear a DVector.
    */
  void Clear(void) {Erase(0, getLength());}

  /**  Replace each element of the vector with it's complex conjugate.
    *  @memo Conjugate a DVector.
    */
  virtual void Conjugate(void) = 0;

  /**  The data vector is extended to the specified length. New data words
    *  are filled with zeros. If the current data length is greater than the
    *  requested length, no action is taken (i.e. the vector is not shortened).
    *  Use ReSize to reduce the vector length or to extend the vector with
    *  uninitialized data.
    *  @memo Extend a DVector.
    *  @param N New size for the data vector.
    */
  virtual void Extend(size_type N) = 0;
};

//======================================  inline DVector methods
inline DVector&
DVector::replace(size_type inx, size_type N, const DVector& v) {
    return replace(inx, N, v, 0, v.getLength());
}

inline void 
DVector::Append(const DVector& data) {
    replace(getLength(), 0, data);
}

inline DVector& 
DVector::insert(size_type inx, const DVector& v, size_type inx2, size_type N) {
    return replace(inx, 0, v, inx2, N);
}

inline double 
DVector::operator*(const DVector& rhs) const {
    return dot(0, rhs, 0, getLength());
}

inline DVector& 
DVector::operator+=(const DVector& rhs) {
    return add(0, rhs, 0, getLength());
}

inline DVector&
DVector::operator-=(const DVector& rhs) {
    return sub(0, rhs, 0, getLength());
}

inline DVector&
DVector::operator+=(double rhs) {
    return bias(0, rhs, getLength()); 
}

inline DVector& 
DVector::operator-=(double rhs) {
    return operator+=(-rhs);
}

inline DVector& 
DVector::operator*=(double rhs) {
    return scale(0, rhs, getLength()); 
}

inline DVector& 
DVector::operator/=(double rhs) {
    return operator*=(1.0/rhs);
} 

inline DVector& 
DVector::operator*=(const DVector& rhs) {
    return mpy(0, rhs, 0, getLength());
} 

inline DVector& 
DVector::operator/=(const DVector& rhs) {
    return div(0, rhs, 0, getLength());
}

#endif  // !def(DVECTOR_HH)
