#ifndef _LIGO_FILTERMODULE_H
#define _LIGO_FILTERMODULE_H
/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: FilterModule						*/
/*                                                         		*/
/* Module Description: Filter module as defined by online system	*/
/*                                                         		*/
/* Revision History:					   		*/
/* Rel   Date     Programmer  	Comments				*/
/* 1.0	 4Aug02   D. Sigg    	First release		   		*/
/*                                                         		*/
/* Documentation References:						*/
/*	Man Pages: FilterModule.html					*/
/*	References: none						*/
/*                                                         		*/
/* Author Information:							*/
/* Name          Telephone       Fax             e-mail 		*/
/* Daniel Sigg   (509) 372-8132  (509) 372-8137  sigg_d@ligo.mit.edu	*/
/*                                                         		*/
/*                                                         		*/
/*                      -------------------                             */
/*                                                         		*/
/*                             LIGO					*/
/*                                                         		*/
/*        THE LASER INTERFEROMETER GRAVITATIONAL WAVE OBSERVATORY.	*/
/*                                                         		*/
/*                     (C) The LIGO Project, 1999.			*/
/*                                                         		*/
/*                                                         		*/
/* Caltech				MIT		   		*/
/* LIGO Project MS 51-33		LIGO Project NW-17 161		*/
/* Pasadena CA 91125			Cambridge MA 01239 		*/
/*                                                         		*/
/* LIGO Hanford Observatory		LIGO Livingston Observatory	*/
/* P.O. Box 1970 S9-02			19100 LIGO Lane Rd.		*/
/* Richland WA 99352			Livingston, LA 70754		*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#include "FilterDesign.hh"
#include "filterwiz/FilterSection.hh"
#include <string>
#include <vector>

namespace filterwiz {


   /// Maximum number of supported filter sections by a filter module.
   const int kMaxFilterSections = 10;
   /// Maximum number of supported SOS by a filter section.
   const int kMaxFilterSOS = 10;


   /// String for filter module defintion
   const char* const kModuleDef = "# MODULES";
   /// String for filter module sampling rate defintion
   const char* const kModuleSamplingDef = "# SAMPLING";
   /// String for filter section design defintion
   const char* const kSectionDesignDef = "# DESIGN";



/** Filter Module Class. Represents a single filter module. Filter
    sections are automatically allocated up to kMaxFilterSections.
   
    @memo Filter Module
    @author Written August 2002 by Daniel Sigg
    @version 1.0
 ************************************************************************/
   class FilterModule {
   public:
      /// Deafult constructor
      FilterModule ();
      /// Constructor
      FilterModule (const char* name, double fsample);
   
      /// Check design of filter sections
      void checkDesign();
   
      /// Get name
      const char* getName() const {
         return fName.c_str(); }
      /// Set name
      void setName (const char* p);
      /// Get sampling rate
      double getFSample() const {
         return fSample; }
      /// Set sampling rate
      void setFSample (double sample);
      /// Guess sampling rate
      void guessFSample (const char* filename);
      /// Access to filter section
      FilterSection& operator[] (int i) {
         return fSect[i]; }
      /// Access to filter section
      const FilterSection& operator[] (int i) const {
         return fSect[i]; }
   
   protected:
      /// Name
      std::string	fName;
      /// Sampling rate
      double		fSample;
      /// List of filter sections
      FilterSectionList	fSect;
   };


/// List of filter modules
   typedef std::vector<FilterModule> FilterModuleList;



}

#endif // _LIGO_FILTERMODULE_H
