#ifndef __EXTENSIONS__
#define __EXTENSIONS__
#endif
#include "laudio.hh"
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef SOLARIS_AUDIO
#include <sys/audioio.h>
#elif defined(OPEN_SOUND_SYSTEM)
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#else
#error 1,"Audio driver not implemented"
#endif
#include <iostream>

using namespace std;

//============================================================================
//
//      File scope utility finctions
//
//============================================================================


//======================================  Convert boolean to "Yes" or "No"
static const char* 
Bool(int yn) {
    return yn ? "Yes" : "No";
}

//===========================================================================
//
//     LAudio class Implementation
//
//===========================================================================

//======================================  Default constructor
LAudio::LAudio(void) 
  : mFD(-1), mOutputSPS(0.0), mDt(0.0), mBias(0.0), mScale(1.0)
{
    int test(1);
    mBigEnd = *(char*)&test == 0;
}

//======================================  Construct an audio I/F
LAudio::LAudio(const char* device)  
  : mFD(-1), mOutputSPS(0.0), mDt(0.0), mBias(0.0), mScale(1.0)
{
    int test(1);
    mBigEnd = *(char*)&test == 0;
    open(device);
}

//======================================  Destroy audio interface
LAudio::~LAudio(void) {
    close();
}

//=======================================  Convert format spec to string
const char*
LAudio::Encode(int enc) const {
    switch (enc) {

#ifdef OPEN_SOUND_SYSTEM
    case AFMT_MU_LAW:
        return "mu-Law";
    case AFMT_A_LAW:
        return "A-Law";
    case AFMT_S16_BE:
    case AFMT_S16_LE:
    case AFMT_U16_BE:
    case AFMT_U16_LE:
      // case AFMT_S32_BE:
      // case AFMT_S32_LE:
    case AFMT_S8:
    case AFMT_U8:
        return "Linear";
#else
    case AUDIO_ENCODING_ULAW:
        return "mu-Law";
    case AUDIO_ENCODING_ALAW:
        return "A-Law";
    case AUDIO_ENCODING_LINEAR:
        return "Linear";
#endif
    }
    return "Unknown";
}

//======================================  Convert to internal format spec.
int
LAudio::Encode(const std::string& enc) const throw(std::runtime_error) 
{
#ifdef OPEN_SOUND_SYSTEM
    if (enc == "mu-Law")   return AFMT_MU_LAW;
    if (enc == "A-Law")    return AFMT_A_LAW;
    if (enc == "Linear")   return mBigEnd ? AFMT_S16_BE : AFMT_S16_LE;
    // if (enc == "Linear32") return mBigEnd ? AFMT_S32_BE : AFMT_S32_LE;
#else
    if (enc == "mu-Law") return AUDIO_ENCODING_ULAW;
    if (enc == "A-Law")  return AUDIO_ENCODING_ALAW;
    if (enc == "Linear") return AUDIO_ENCODING_LINEAR;
#endif
    else throw std::runtime_error("Invalid coding type");
    return 0;
}

//======================================  Get the data encoding
const char*
LAudio::getEncoding(void) const throw(std::runtime_error) {
#ifdef OPEN_SOUND_SYSTEM
    int fmt = AFMT_QUERY;
    if (ioctl(mFD, SNDCTL_DSP_SETFMT, &fmt) < 0) {
        throw std::runtime_error("Unable to query format");
    }
    return Encode(fmt);
#else
    audio_info_t info;
    getStatus(&info);    
    return Encode(info.play.encoding);
#endif
}

//======================================  Get the monitor gain.
LAudio::count_t
LAudio::getMonitorGain(void) const throw(std::runtime_error) {
#ifdef OPEN_SOUND_SYSTEM
    return 0;
#else
    audio_info_t info;
    getStatus(&info);
    return info.monitor_gain;
#endif
}

//======================================  Get the data precision.
LAudio::count_t
LAudio::getPrecision(void) const throw(std::runtime_error) {
#ifdef OPEN_SOUND_SYSTEM
    int fmt = AFMT_QUERY;
    if (ioctl(mFD, SNDCTL_DSP_SETFMT, &fmt) < 0) {
        throw std::runtime_error("Unable to query format");
    }
    int prec = -1;
    switch (fmt) {
    case AFMT_A_LAW:
    case AFMT_MU_LAW:
    case AFMT_S8:
    case AFMT_U8:
        prec = 8;
	break;
    case AFMT_S16_BE:
    case AFMT_S16_LE:
    case AFMT_U16_BE:
    case AFMT_U16_LE:
        prec = 16;
	break;
	// case AFMT_S32_BE:
	// case AFMT_S32_LE:
        // prec = 32;
	// break;
    default:
        throw std::runtime_error("Unknown data format");
    };
    return prec;
#else
    audio_info_t info;
    getStatus(&info);
    return info.play.precision;
#endif
}


//======================================  get interface status (Solaris only)
#ifdef SOLARIS_AUDIO
void
LAudio::getStatus(audio_info_t* info) const throw(std::runtime_error) {
    if (ioctl(mFD, AUDIO_GETINFO, info) < 0) {
        std::string error = "Error getting status: ";
	error += strerror(errno);
	throw std::runtime_error(error);
    }
}
#endif

//=======================================  Get the sample rate.
LAudio::count_t
LAudio::getSpeed(void) const throw(std::runtime_error) {
    count_t Rate=0;
#ifdef OPEN_SOUND_SYSTEM
    if (ioctl(mFD, SNDCTL_DSP_SPEED, &Rate) < 0) {
        throw std::runtime_error("Unable to request sample rate");
    }
#else
    audio_info_t info;
    getStatus(&info);
    if (mMode == m_listen) Rate = info.record.sample_rate;
    else                   Rate = info.record.sample_rate;
#endif
    return Rate;
}

//=======================================  Get the output loudness level.
double
LAudio::getVolume(void) const throw(std::runtime_error) {
#ifdef SOLARIS_AUDIO
    audio_info_t info;
    getStatus(&info);
    return double(info.play.gain)/AUDIO_MAX_GAIN;
#else
    return 0.0;
#endif
}

//=======================================  Set data encoding
void
LAudio::setEncoding(const char* code) throw(std::runtime_error) {
    int fmt = Encode(code);
#ifdef OPEN_SOUND_SYSTEM
    if (ioctl(mFD, SNDCTL_DSP_SETFMT, &fmt) < 0) {
        throw std::runtime_error("Unable to set format");
    }
#else
    audio_info_t info;
    getStatus(&info);
    if (mMode != m_sing)   info.record.encoding = fmt;
    if (mMode != m_listen) info.play.encoding   = fmt;
    setStatus(&info);
#endif
    if (mMode != m_sing)   mInStat.format  = fmt;
    if (mMode != m_listen) mOutStat.format = fmt;
}


//=======================================  Set gain for monitor output
void
LAudio::setMonitorGain(count_t gain) throw(std::runtime_error) {
#ifndef OPEN_SOUND_SYSTEM
    audio_info_t info;
    if (mMode == m_monitor) {
        getStatus(&info);
	info.record.encoding = gain;
	setStatus(&info);
    }
#endif
}

//========================================  Select I/O ports
void
LAudio::setPorts(count_t mode) throw(std::runtime_error) {
#ifndef OPEN_SOUND_SYSTEM
    audio_info_t info;
    getStatus(&info);
    info.play.port = mode;
    setStatus(&info);
#endif
}

//======================================  Drain device buffers.
void
LAudio::drain(void) throw(std::runtime_error) {

#ifdef OPEN_SOUND_SYSTEM
    int opcd = SNDCTL_DSP_SYNC;
#else
    int opcd = AUDIO_DRAIN;
#endif
    if (::ioctl(mFD, opcd, 0) < 0) {
        throw std::runtime_error("Error draining audio buffer");
    }
}

//======================================  Mute the output
void
LAudio::mute(bool mute) throw(std::runtime_error) {
#ifndef OPEN_SOUND_SYSTEM
    audio_info_t info;
    getStatus(&info);
    if (mute) info.output_muted = 1;
    else      info.output_muted = 0;
    setStatus(&info);
#endif
}

//======================================  Set the linear input transformation
void 
LAudio::setInputTransform(double bias, double scale) {
    mBias  = bias;
    mScale = scale;
}

//======================================  Set the number of bits per sample
void
LAudio::setPrecision(count_t bits) throw(std::runtime_error) {
    int fmt;
#ifdef OPEN_SOUND_SYSTEM
    if (bits == 8) {
        fmt = AFMT_U8;
    } else if (bits == 16) {
        fmt = mBigEnd ? AFMT_S16_BE : AFMT_S16_LE;
    // } else if (bits == 32) {
    //     fmt = mBigEnd ? AFMT_S32_BE : AFMT_S32_LE;
    } else {
        throw("Unsupported precision requested");
    }
    if (ioctl(mFD, SNDCTL_DSP_SETFMT, &fmt) < 0) {
        throw("Unable to set format");
    }
#else
    audio_info_t info;
    getStatus(&info);
    fmt = AUDIO_ENCODING_LINEAR;
    if (mMode != m_sing) {
        info.record.precision = bits;
	info.record.encoding  = fmt;
    }
    if (mMode != m_listen) {
        info.play.precision = bits;
	info.play.encoding  = fmt;
    }
    setStatus(&info);
#endif
    if (mMode != m_sing) {
        mInStat.precision = bits;
	mInStat.format    = fmt;
    }
    if (mMode != m_listen) {
        mOutStat.precision = bits;
	mOutStat.format    = fmt;
    }
}

//====================================== Set the sample rate
void 
LAudio::setSpeed(count_t Rate) throw(std::runtime_error) {
#ifdef OPEN_SOUND_SYSTEM
    if (ioctl(mFD, SNDCTL_DSP_SPEED, &Rate) < 0) {
        throw std::runtime_error("Unable to set requested sample rate");
    }
#else
    audio_info_t info;
    getStatus(&info);
    if (mMode != m_listen) info.play.sample_rate   = Rate;
    if (mMode != m_sing)   info.record.sample_rate = Rate;
    setStatus(&info);
#endif
    if (mMode != m_listen) {
        mOutStat.sample_rate = Rate;
	mOutputSPS = double(Rate);
    }
    if (mMode != m_sing)   mInStat.sample_rate  = Rate;
}

//======================================  Set the device status (Solaris only)
#ifdef SOLARIS_AUDIO
void 
LAudio::setStatus(const audio_info_t* info) throw(std::runtime_error) {
    drain();
    if (::ioctl(mFD, AUDIO_SETINFO, info) < 0) {
        throw std::runtime_error("Error setting status");
    }
}
#endif

//====================================== Set the loudness
void 
LAudio::setVolume(double volume) throw(std::runtime_error) {
#ifdef SOLARIS_AUDIO
    audio_info_t info;
    getStatus(&info);
    info.play.gain = int(volume*AUDIO_MAX_GAIN);
    setStatus(&info);
#endif
}

//======================================  Dump out the status
void 
LAudio::dumpStatus(std::ostream& out) const {
#ifdef SOLARIS_AUDIO
    audio_info_t info;
    try {
        getStatus(&info);
        out << "Device type: " << getDevice() << std::endl;
        dumpChannel(out, "Record channel:", info.record);
        dumpChannel(out, "Play channel:", info.play);
	out << "Monitor Gain: " << info.monitor_gain << std::endl;
	out << "Output Muted: " << Bool(info.output_muted) << std::endl;
    } catch (std::runtime_error& e) {
        out << e.what() << std::endl;
    }
#endif
}

//======================================  Dump data from a channel.
#ifdef SOLARIS_AUDIO
void 
LAudio::dumpChannel(std::ostream& out, const char* title, 
		    const audio_prinfo_t& info) const throw(std::runtime_error)
{
    out << title << std::endl;
    out << "    Sample rate:     " << info.sample_rate << std::endl;
    out << "    Channels:        " << info.channels << std::endl;
    out << "    Precision:       " << info.precision << std::endl;
    out << "    Encoding:        " << Encode(info.encoding) << std::endl;
    out << "    Gain:            " << info.gain << std::endl;
    out << "    Port:            " << info.port << std::endl;
    out << "    Buffer size:     " << info.buffer_size << std::endl;
    out << "    Samples:         " << info.samples << std::endl;
    out << "    EOF:             " << info.eof << std::endl;
    out << "    Pause:           " << Bool(info.pause) << std::endl;
    out << "    Error:           " << Bool(info.error) << std::endl;
    out << "    Waiting:         " << Bool(info.waiting) << std::endl;
    out << "    Balance:         " << int(info.balance) << std::endl;
    out << "    Open:            " << Bool(info.open) << std::endl;
    out << "    Active:          " << Bool(info.active) << std::endl;
    out << "    Available Ports: " << info.avail_ports << std::endl;
    out << std::endl;
}
#endif

//======================================  Get the device name string.
std::string 
LAudio::getDevice(void) const {
#ifdef SOLARIS_AUDIO
    audio_device_t devtype;
    if (! ::ioctl(mFD, AUDIO_GETDEV, &devtype)) {
	std::string name = devtype.name;
	name += "-";
	name += devtype.version;
	name += "-";
	name += devtype.config;
	return name;
    } else {
        return "Unknown";
    }
#else
    return "Unknown";
#endif
}

//=======================================  Test if channel is open
bool 
LAudio::isOpen(void) const {
    return (mFD >= 0);
}

//=======================================  Open a device
void 
LAudio::open(const char* device, iomode mode) {

    //----------------------------------  Get a default name if needed
    if (!device || !*device) device = ::getenv("AUDIODEV");
    if (!device || !*device) device = "/dev/audio";

    //----------------------------------  Open the device.
    mMode = mode;
    switch (mMode) {
    case m_sing:
        mFD = ::open(device, O_WRONLY);
	break;
    case m_listen:
    case m_monitor:
        mFD = ::open(device, O_RDONLY);
	break;
    case m_converse:
        mFD = ::open(device, 0);
        break;
    }

    //----------------------------------  Print an error jic.
    if (mFD < 0) {
        std::string error = "Unable to open device ";
	error += device;
	error += " Error: ";
	error += strerror(errno);
	throw std::runtime_error(error);
    }

#ifdef SOLARIS_AUDIO
    audio_info_t info;
    //------------------------------  Get the status, save some fields.
    getStatus(&info);
    mOutputSPS = double(info.play.sample_rate);

    //------------------------------  Mode dependent initialization.
    switch (mMode) {
    case m_sing:
        info.play.encoding    = Encode("Linear");
	info.play.precision   = 16;
	break;
    case m_listen:
        info.record.encoding  = Encode("Linear");
	info.record.precision = 16;
	break;
    case m_monitor:
        info.play.encoding    = Encode("Linear");
	info.play.precision   = 16;
	info.record.encoding  = Encode("Linear");
	info.record.precision = 16;
	info.monitor_gain     = 128;
	break;
    case m_converse:
        info.play.encoding    = Encode("Linear");
	info.play.precision   = 16;
	info.record.encoding  = Encode("Linear");
	info.record.precision = 16;
	info.monitor_gain     = 0;
	break;
    }
    setStatus(&info);
#else
    mOutputSPS = getSpeed();
    setPrecision(16);
#endif
}


//======================================  Close an opened device.
void 
LAudio::close(void) {
    if (mFD < 0) return;
    if (mMode != m_listen) drain();
    ::close(mFD);
    mFD = -1;
}

//======================================  Output a stretch of sound.
void 
LAudio::sing(short data[], count_t length) {
    int rc = ::write(mFD, data, length*sizeof(short));
    if (rc < 0) perror("Error in sing");
    else if (rc < length) cerr << "Partial write: " << rc << "/" << length 
			       << " bytes written" << std::endl;
}

//======================================  Resample data and output it.
void 
LAudio::reSample(double freq, short data[], count_t length) {

    //----------------------------------  Get the number of audio samples
    double inPeriod = 1./freq;
    count_t nReSample = int((inPeriod*double(length-1)-mDt)*mOutputSPS);

    //----------------------------------  Buy some temporary storage.
    short* rData = new short[nReSample];

    //----------------------------------  Calculate the storage
    for (count_t i=0 ; i<nReSample ; i++) {
        double tau = double(i)/mOutputSPS + mDt;
	if (tau < 0.0) {
	    rData[i] = transform((mOutputVal-data[0])*tau/mDt + data[0]);
	} else {
	    double  index = tau * freq;
	    count_t floor = count_t(index);
	    index -= floor;
	    rData[i] = transform((1.-index)*data[floor] + index*data[floor+1]);
	}
    }
    mDt += double(nReSample)/mOutputSPS - double(length)*inPeriod;
    mOutputVal = data[length-1];
    sing(rData, nReSample);
    delete[] rData;
}

//======================================  Resample data and output it.
void 
LAudio::reSample(double freq, float data[], count_t length) {

    //----------------------------------  Get the number of audio samples
    double inPeriod = 1./freq;
    count_t nReSample = int((inPeriod*double(length-1)-mDt)*mOutputSPS);

    //----------------------------------  Buy some temporary storage.
    short* rData = new short[nReSample];

    //----------------------------------  Calculate the storage
    for (count_t i=0 ; i<nReSample ; i++) {
        double tau = double(i)/mOutputSPS + mDt;
	if (tau < 0.0) {
	    rData[i] = short(mOutputVal*tau/mDt 
			     +transform(data[0])*(1.-tau/mDt));
	} else {
	    double index = tau * freq;
	    count_t floor = count_t(index);
	    index -= floor;
	    rData[i] = transform((1.-index)*data[floor] + index*data[floor+1]);
	}
    }
    mDt += double(nReSample)/mOutputSPS - double(length)*inPeriod;
    mOutputVal = transform(data[length-1]);

    sing(rData, nReSample);
    delete[] rData;
}

LAudio::DspStatus::DspStatus(void) 
  : valid(false), sample_rate(0), format(-1), precision(0), buffer_size(0),
    gain(-1), samples(0)
{}
