#ifndef DATAMSG_HH
#define DATAMSG_HH

#include "lmsg/Message.hh"

namespace lmsg {
    class TransInput;
    class TransOutput;

  /**  Template for messages with a single data item. The data item may 
    *  be any defined class with the following properties.
    *  \begin{itemize}
    *  \item A copy constructor
    *  \item An #operator>>(TransInput&, T&)# defined either intrinsically 
    *  in the #TransInput# class or as an external function.
    *  \item An #operator<<(TransOutput&, const T&)# defined either 
    *  intrinsically in the #TransOutput# class or as an external function.
    *  \end{itemize}
    *  @memo Data message template.
    *  @author John Zweizig
    *  @version $Revision: 1.2 $; Modified $Date: 2003/04/10 18:11:23 $
    */
  template<class T, unsigned int MsgID>
  class DataMsg : public Message {
  public:

    /**  Construct a message.
      *  @memo Default constructor.
      */
    DataMsg(void) {}

    /**  Construct a message identical to an existing message.
      *  @memo Copy constructor.
      */
    DataMsg(const DataMsg<T,MsgID>& x) : mData(x.mData) {}

    /**  Construct a message and initialize data to a specified value.
      *  @memo Data constructor.
      */
    explicit DataMsg(const T& x) : mData(x) {}

    /**  Destroy a message.
      *  @memo Default destructor.
      */
    virtual ~DataMsg(void) {}

    /**  Get the message text length.
      *  @memo Text length.
      *  @return Always zero for dataless messages.
      */
    size_type size(void) const {return 0;}

    /**  Write the message data to translation stream.
      *  @memo Get message data.
      */
    void getData(TransOutput& out) const;

    /**  Get the message type.
      *  @memo Message type.
      *  @return As specified by template argument.
      */
    MsgHeader::MsgType getType(void) const {return MsgID;}

    /**  Read data from a transltion stream into the message.
      *  @memo Set message data.
      */
    void setData(TransInput& in) throw(std::runtime_error);

    /**  Get a reference to the data portion of the message.
     */
    T& refData(void) {return mData;}
    const T& refData(void) const {return mData;}

  private:
    T mData;
  };

}  // namespace lmsg

#endif   // DATAMSG_HH
