#ifndef METIOWRAP_HH
#define METIOWRAP_HH

struct MetaioParseEnvironment;

#include <string>

namespace xsil {

  /**  MetaIO is a wrapper for the LigoTools metaio package. The package is used
    *  by various tools to read (or write?) Ligo light-weight tables.
    *  @memo    I/O to/from a ligo lightweight table.
    *  @version 1.0; Last modified February 19, 2004
    *  @author  John Zweizig
    */
  class MetaIO {
  public:
    /**  Construct a MetaIO reader.
      */
    MetaIO(void);

    /**  Construct a ligo-lightweight table reader.
      */
    MetaIO(const char* file, const char* table=0);

    /**  Destroy a ligo-lightweight table reader.
      */
    ~MetaIO(void);

    /**  Close a Ligo-lightweight reader.
      *  @return zero if successful.
      */
    int close(void);

    const char* getColumnName(int icol) const;

    /**  Get a floating point number from the specified column of the current
      *  row. If the column data type differs from the requested data type
      *  GetFloat attempts to convert the data to a double.
      */
    double getFloat(const char* cname, double def=0.0) const;

    /**  Get a long integer number from the specified column of the current
      *  row.If the column data type differs from the requested data type
      *  GetInt attempts to convert the data to a long integer.
      */
    long getInt(const char* cname, long def=0) const;

    int getNColumn(void) const;

    /**  Read the next row of the table from the opened file.
      *  @return Zero if successful.
      */
    int getRow(void);

    /**  Get a string value from the specified column of the current
      *  row. If the column data type differs from the requested data type
      *  GetString attempts to convert the data to a string.
      */
    std::string getString(const char* cname, const char* def="") const;

    /**  Get the table name from the header information.
     */
    const char* getTableName(void) const;

    /**  Get the column type name.
     */
    const char* getTypeName(int icol) const;

    /** Test whether the parse environment is ope. Returns true if the input 
      * file is attached and open.
      */
    bool is_open(void) const;

    /**  The specified file is opened and the header information is read.
      *  @param  file Name of xml file to be processed.
      *  @return Zero indicates success.
      *  @memo   Open an xml table file.
      */
    int open(const char* file, const char* table=0);

  private:
    MetaioParseEnvironment* mParseEnv;
  };

  inline bool
  MetaIO::is_open(void) const {
      return mParseEnv != 0;
  }

} // namespace xsil

#endif // METTAIOWRAP_HH
