#ifndef XSIL_METATABLE_HH
#define XSIL_METATABLE_HH

#include "xsil/table.hh"
#include "xsil/stream.hh"
#include <string>
#include <list>

namespace xsil {

  /**   Specialized table with automatic filling and reference capabilities.
    */
  class MetaTable : public table {

    //----------------------------------  Class to hold data pointers
    class eventcolumn {
    public:
      enum d_type {
	t_none,
	t_chars,
	t_charu,
	t_string,
	t_int4s,
	t_longs,
	t_real4,
	t_real8,
	t_cite
      };
      eventcolumn(void);
      eventcolumn(d_type t, const void* a, const char* p=0);
      eventcolumn(const eventcolumn& x);
      ~eventcolumn(void);
      d_type      data_Type;
      const void* data_Addr;
      std::string data_Param;
    };

    //----------------------------------  Column list data type
    typedef std::list<eventcolumn> ColumnList;
    typedef ColumnList::const_iterator col_iter;

  public:

    /**  Construct an empty Meta-table.
      *  Empty table constructor.
      */
    MetaTable(void);

    /**  Construct a Named Meta-table.
      *  @memo Name meta-table constructor.
      */
    MetaTable(const char* Name, const char* Type=0);

    /**  Construct a Named Meta-table.
      *  @memo Name meta-table constructor.
      */
    MetaTable(const MetaTable& x);

    /**  Create an identical copy og the meta-table and return a pointer to
      *  the copy.
      *  @memo Clone a Meta-table.
      */
    // MetaTable* Clone(void) const;

    /**  Destroy a meta-table.
      *  @memo Meta-table destructor.
      */
    virtual ~MetaTable(void);

    /**  Dump out the column data.
      *  @memo Clear metatable.
      */
    void check(const char* title, int csize=0) const;

    /**  Clear all table column data.
      *  @memo Clear metatable.
      */
    void clear(void);

    /**  Specify a table column. The column name and type are included.
      *  an address field specifies a static address from which the data 
      *  will be copied.
      *  @memo Add a column to the table.
      *  @param Name Column name attribute field.
      *  @param Addr Pointer to the data field.
      */
    void defineColumn(const char* Name, const int*         Addr);
    void defineColumn(const char* Name, const long*        Addr);
    void defineColumn(const char* Name, const float*       Addr);
    void defineColumn(const char* Name, const double*      Addr);
    void defineColumn(const char* Name, const char*        Addr,
		      bool ilwc=false);
    void defineColumn(const char* Name, const std::string* Addr, 
		      bool ilwc=false);
    void defineColumn(const char* Name, const UCVec*       Addr);
    void defineColumn(const char* Name, const MetaTable*   Addr, 
		      const char* Col=0);

    /**  Get the number of data columns that have already been defined.
      *  @memo Get the number of columns.
      */
    int getNColumn(void) const;

    /**  Get the number of data rows that have already been defined.
      *  @memo Get the number of rows.
      */
    int getNRow(void) const;

    /**  Copy all specified data fields (one for each table column) into the 
      *  data stream.
      *  @memo Copy an event into the data stream.
      */
    void putRow(void);

    /**  Delete the contens of the table data stream.
      *  @memo Reset the data stream.
      */
    void resetStream(void);

    /**  Return a reference string to the last row of the table. The 
      *  citation is of the form #<table>:<column>:<row>#.
      *  @memo Get a reference to the last row of the table.
      */
    std::string citeTable(const std::string& col) const;

    /**  Return a reference string to the specified row of the table. The 
      *  citation is of the form #<table>:<column>:<row>#. No validity check
      *  is performed on either the column name or the row number.
      *  @memo Get a reference to the last row of the table.
      */
    std::string citeTable(const std::string& col, int row) const;

  private:
    std::string getTableName(void) const;
    std::string getItemName(const std::string& x) const;

  private:
    std::string mGroup;
    ColumnList  mCList;
    int         mRow;
  };

  //====================================  Inline row and column counts.
  inline int
  MetaTable::getNRow(void) const {
      return mRow;
  }

  inline int
  MetaTable::getNColumn(void) const {
      return mCList.size();
  }

}  // namespace xsil

#endif  //  XSIL_METATABLE_HH
