#ifndef SEGLIST_HH
#define SEGLIST_HH

#include <iosfwd>
#include <stdexcept>
#include "TSeries.hh"
#include <vector>

class LockSegment {
public:
  typedef unsigned long flag_type;
  typedef long          id_type;
  LockSegment(id_type id, const Time& t0, const Interval dT, flag_type flg=0);
  Interval getDuration(void) const;
  Time getEndTime(void) const;
  flag_type getFlags(void) const;
  id_type getID(void) const;
  const Time& getStartTime(void) const;
  bool inSegment(const Time& t) const;
  bool inSegment(const Time& t, const Time& t2) const;
  bool operator< (const LockSegment& s) const;
  bool operator< (const Time& s) const;
  void setRange(const Time& t0, Interval dT);

private:
  id_type   mID;
  Time      mStart;
  Interval  mDuration;
  flag_type mFlags;
};

//======================================  inline methods
inline Interval
LockSegment::getDuration(void) const {
    return mDuration;
}

inline Time 
LockSegment::getEndTime(void) const {
    return mStart+mDuration;
}

inline LockSegment::flag_type
LockSegment::getFlags(void) const {
    return mFlags;
}

inline LockSegment::id_type
LockSegment::getID(void) const {
    return mID;
}

inline const Time& 
LockSegment::getStartTime(void) const {
    return mStart;
}

inline bool 
LockSegment::inSegment(const Time& t) const {
    return ((t >= mStart) && (t < getEndTime()));
}

inline bool 
LockSegment::inSegment(const Time& t1, const Time& t2) const {
    return ((t1 < t2) && (t1 >= mStart) && (t2 <= getEndTime()));
}

inline bool 
LockSegment::operator< (const LockSegment& s) const {
    return mStart < s.mStart;
}

//======================================  Segment list class
#define SEGMENT_TINF (0x7fffffff)
class LockSegList {
public:
  typedef unsigned long size_type;
public:
  explicit LockSegList(const char* lid, const char* file=0);
  virtual ~LockSegList(void) {}

  /**  Erase all the segments in the list.
    */
  void clear(void);

  /**  Look for the segment containing the specified time.
    */
  int find(const Time& t) const throw(std::domain_error);

  /**  Find the first segment that ends after the specified time.
    *  The index of the found segment is returned. If ther are no known 
    *  segments after the specified time, an index equal to the size of 
    *  segment list is returned.
    */ 
  int findafter(const Time& t) const;

  const char* getListID(void) const;

  bool inSegment(const Time& t) const;

  bool inSegment(const Time& t, const Time& t2) const;

  void insert(const LockSegment& l);

  void invert(void);

  void read(const std::string& file);

  void setDebug(int lvl);
  size_type size(void) const;

  const LockSegment& operator[](size_type seg) const;

  const LockSegment& operator[](const Time& t) const throw(std::domain_error);

  LockSegList& operator|=(const LockSegList& t);

  LockSegList& operator&=(const LockSegList& t);

  std::ostream& putID(std::ostream& out, size_type inx) const;

private:
  LockSegment& ref(size_type seg);

public:
  typedef std::vector<LockSegment> SegList_type;
  typedef SegList_type::const_iterator seg_iter;

private:
  std::string  mListID;
  SegList_type mList;
  int          mDebug;
};

std::ostream&
operator<<(std::ostream& o, const LockSegment& s);

//======================================  Segment list inlined methods
inline const char* 
LockSegList::getListID(void) const {
    return mListID.c_str();
}

inline const LockSegment&
LockSegList::operator[](size_type seg) const {
    return mList[seg];
}

inline const LockSegment&
LockSegList::operator[](const Time& t) const throw(std::domain_error) {
    return mList[find(t)];
}

inline LockSegment&
LockSegList::ref(size_type seg) {
    return mList[seg];
}

inline LockSegList::size_type
LockSegList::size(void) const {
    return mList.size();
}

#endif  // !defined(SEGLIST_HH)
