/***************************************************************************
    File        : NoiseFloorMonitor.cpp
    Description : Implements the NoiseFloorMonitor
 ---------------------------------------------------------------------------
    Begin       : Tue March 30 2005
    Author(s)   : Roberto Grosso
 ***************************************************************************/


#include "NoiseFloorMonitor.h"

//===================================== Auxiliary functions
inline void Estimator(const std::vector<double>& ts,double& mean,double& stdDev);
inline void Normalize(std::vector<double>& data);
inline void MixWithOscilator(const double carrierFreq,const double sampleRate,std::vector<double>& signal);
inline void GetTrend(const double sigma,std::vector<double>& ts,double& maxVal,double& percentage);

//===================================== generate the main routine
EXECDAT(NoiseFloorMonitor)

//===================================== Local help functions
void ReadNotchFilterFreq(const std::string file,std::vector<double>& freq,std::vector<double>& freqbw);



//======================================  NoiseFloorMonitor constructor
NoiseFloorMonitor::NoiseFloorMonitor(int argc, const char *argv[]) : DatEnv(argc, argv), MonServer("NoiseFloorMon")
{
  //----------------------------------  Process arguments
  //--------------------------------------------------------------------
  std::string prFile;
  std::string fqFile;

  // Process input parameters
  for (int i=1 ; i<argc ; i++)
  {
    std::string argi = argv[i];
    if (argi == "-param") {
      prFile = argv[++i];
    } else if (argi == "-freq") {
      fqFile = argv[++i];
    } else if (argi == "-inlist") {
      ++i;
    } else if (argi == "-maxframes") {
      ++i;
    } else if (argi == "-debug") {
      ++i;
    } else {
      std::string strError = std::string("NoiseFloorMonitor() Unrecognized argument: ") + std::string(argi);
      gwd::Error(strError);
    }
  }

  //----------------------------------  init class variables
  //--------------------------------------------------------------------
  mCounterTimeStrides  = 0;
  mCounterWitheFilter  = 0;
  mTStride             = 0.;

  mOrderNotchFilter = 0;
  mOrderWhiteFilter = 0;
  mNoOfFreqPsd      = 0;

  mSizePsd                 = 0.;
  mRMSizeForWhiteningFlt   = 0.;
  mWindowSizeRunningMedian = 0.;
  mSampleRate              = 0.;

  //----------------------------------  read parameters
  //--------------------------------------------------------------------
  gwd::Parameter param;
  param.ReadParameters(prFile);

  //----------------------------------  set the time stride.
  mTStride = param.TimeStride();
  getDacc().setStride(param.TimeStride());
  mTStrideSize = mTStride.GetSecs();

  //----------------------------------  Specify channel to scan
  //--------------------------------------------------------------------
  getDacc().addChannel(param.DetectorChannel());
  getDacc().setIgnoreMissingChannel(true);

  //----------------------------------  copy parameters
  //--------------------------------------------------------------------
  mOrderNotchFilter = param.OrderNotchFilter();
  mOrderWhiteFilter = param.OrderWhiteFilter();
  mNoOfFreqPsd      = param.NoOfFreqPSD();
  mSizePsd          = param.SizeForPSD();
  mRMSizeForWhiteningFlt   = param.RunningMedianSizeForWhiteningFilter();
  mWindowSizeRunningMedian = param.WindowSizeRunningMedian();
  mDetectorChannel  = param.DetectorChannel();

  mFltOrder        = param.FltOrder();
  mUpperCutoffFreq = param.UpperCutoffFreq();
  mBandFltOrder    = param.BandFltOrder();
  mBandFreqA       = param.BandFreqA();
  mBandFreqB       = param.BandFreqB();
  mBandFreqC       = param.BandFreqC();
  mBandFreqD       = mUpperCutoffFreq;

  mUpdateWhiteFlt = param.UpdateWhiteFilter();

  mLockConditionName = mDetectorChannel.substr(0,2) + std::string(":Both_arms_locked_strict_cm");

  //----------------------------------  oversampling factor for resampling and decimation
  //-------------------------------------------------------------------------------------
  mOverSamplingFactor = 1.2;

  //----------------------------------  sample rate is equal twice of cutoff frequency (Nyquist)
  //--------------------------------------------------------------------------------------------
  mSampleRate = 2.*mUpperCutoffFreq;

  //----------------------------------  assume signal sampling rate
  //--------------------------------------------------------------------
  mSignalSampleRate = double(16384);

  //----------------------------------  sample rate must be less than or equal to signal sampling rate
  //--------------------------------------------------------------------------------------------------
  if (mSampleRate > mSignalSampleRate)
    mSampleRate = mSignalSampleRate;

  //----------------------------------  read frequencies for the Notch filter
  //-------------------------------------------------------------------------
  ReadNotchFilterFreq(fqFile,mFreqNotchFilter,mFreqBndWidthNotFilter);

  //----------------------------------  compute total filter delay
  //--------------------------------------------------------------------
  // cutoff filter operates at mSignalSampleRate
  mFilterDelay = static_cast<double>(mFltOrder)/mSignalSampleRate;
  // add delay due to whitening filter at mSampleRate
  mFilterDelay += static_cast<double>(mOrderWhiteFilter)/mSampleRate;
  // add delay due to Notch filter at mSampleRate
  mFilterDelay += static_cast<double>(mOrderNotchFilter)/mSampleRate;
  // add delay due to low-pass or band-pass or high-pass filter at mSampleRate
  mFilterDelay += static_cast<double>(mBandFltOrder)/mSampleRate;

  //----------------------------------  compute buffer for time continuous processing
  //---------------------------------------------------------------------------------
  // Create a buffer which is as long as the filter dealy
  // and the running median filter plus a time shift. The
  // running median cut seconds from the tail of the time
  // series.
  mBufferTimeShift = 0.5; // in sec
  unsigned int buffSize = static_cast<unsigned int>((mBufferTimeShift + mFilterDelay + mWindowSizeRunningMedian)*mSignalSampleRate);
  buffer.resize(buffSize);
  std::fill(buffer.begin(),buffer.end(),double(0));

  // Check that buffer size is shorter the time stride
  if ((mBufferTimeShift + mFilterDelay + mWindowSizeRunningMedian) > mTStrideSize)
  {
    std::string strError = std::string("time stride shorter than running median window size and filter delay\n")
                         + std::string("set time stride >= running median window + filter delay + 1 sec.");
    gwd::Error(strError);
  }

  //----------------------------------  check if time stride is large enough
  //--------------------------------------------------------------------
  const unsigned int wSize    = (unsigned int) (mRMSizeForWhiteningFlt);
  const unsigned int segSize  = (unsigned int) (mSampleRate * mSizePsd) + wSize;
  const unsigned int tshift   = (unsigned int) (5. * mSampleRate)+ buffer.size(); // shift by 5 secs. + buffer size
  const unsigned int tsSize   = (unsigned int) (mTStrideSize*mSampleRate);
  if ((tshift+segSize) >= tsSize)
  {
    std::string strError = std::string("The size if sec. of the time series used for calculating the psd is too large\n")
                         + std::string("The size should be psd size  < TimeStrid - 5 sec.");
    gwd::Error(strError);
  }

  //---------------------------------- compute frequecy pass filters
  //--------------------------------------------------------------------
  gwd::LowPassFilter  lpf;
  gwd::BandPassFilter bpf;
  gwd::HighPassFilter hpf;

  //----------------------------------  compute the cutoff filter
  //--------------------------------------------------------------------
  lpf.ComputeFilter(mSignalSampleRate,mUpperCutoffFreq,mFltOrder,mCutoffFlt);

  //----------------------------------  compute the low-pass filter
  //--------------------------------------------------------------------
  lpf.ComputeFilter(mSampleRate,mBandFreqA,mBandFltOrder,mLpFltA);

  //----------------------------------  compute the band-pass filter
  //--------------------------------------------------------------------
  bpf.ComputeFilter(mSampleRate,mBandFreqA,mBandFreqB,mBandFltOrder,mBpFltB);
  bpf.ComputeFilter(mSampleRate,mBandFreqB,mBandFreqC,mBandFltOrder,mBpFltC);

  //----------------------------------  compute the high-pass filter
  //--------------------------------------------------------------------
  hpf.ComputeFilter(mSampleRate,mBandFreqC,mBandFltOrder,mHpFltD);

  //----------------------------------  compute the low-pass filter
  //--------------------------------------------------------------------
  lpf.ComputeFilter(mSampleRate,mBandFreqB-mBandFreqA,mBandFltOrder,mLpFltB);
  lpf.ComputeFilter(mSampleRate,mBandFreqC-mBandFreqB,mBandFltOrder,mLpFltC);
  lpf.ComputeFilter(mSampleRate,mBandFreqD-mBandFreqC,mBandFltOrder,mLpFltD);

  //----------------------------------  compute decimate factors and sample rates
  //----------------------------------  band A
  //--------------------------------------------------------------------
  mSampleRateBandA = 2.*mOverSamplingFactor*mBandFreqA;
  mDecimateFactorA = (int)std::floor(mSampleRate/mSampleRateBandA);
  if (mDecimateFactorA < 1) mDecimateFactorA = 1;
  mSampleRateBandA = mSampleRate / static_cast<double>(mDecimateFactorA);
  //----------------------------------  band B
  //--------------------------------------------------------------------
  double bandwidth = mBandFreqB-mBandFreqA;
  mSampleRateBandB = 2.*mOverSamplingFactor*bandwidth;
  mDecimateFactorB = (int)std::floor(mSampleRate/mSampleRateBandB);
  if (mDecimateFactorB < 1) mDecimateFactorB = 1;
  mSampleRateBandB = mSampleRate / static_cast<double>(mDecimateFactorB);
  //----------------------------------  band C
  //--------------------------------------------------------------------
  bandwidth = mBandFreqC-mBandFreqB;
  mSampleRateBandC = 2.*mOverSamplingFactor*bandwidth;
  mDecimateFactorC = (int)std::floor(mSampleRate/mSampleRateBandC);
  if (mDecimateFactorC < 1) mDecimateFactorC = 1;
  mSampleRateBandC = mSampleRate / static_cast<double>(mDecimateFactorC);
  //----------------------------------  band D
  //--------------------------------------------------------------------
  bandwidth = mBandFreqD-mBandFreqC;
  mSampleRateBandD = 2.*mOverSamplingFactor*bandwidth;
  mDecimateFactorD = (int)std::floor(mSampleRate/mSampleRateBandD);
  if (mDecimateFactorD < 1) mDecimateFactorD = 1;
  mSampleRateBandD = mSampleRate / static_cast<double>(mDecimateFactorD);

  //----------------------------------  compute Notch filter
  //--------------------------------------------------------------------
  //  std::cout << " ... compute notch filter" << std::endl;
  gwd::NotchFilter nf;
  nf.ComputeFilter(mSampleRate,mOrderNotchFilter,mFreqNotchFilter,mFreqBndWidthNotFilter,mNotchFilter);

  //----------------------------------  estimate threshold with Gaussian White noise
  //--------------------------------------------------------------------
  EstimateThreshold();

  //---------------------------------- Specify time series names
  //--------------------------------------------------------------------
  std::ostringstream ou1;
  ou1 << "NF-0-" << ((int)mBandFreqA) << "-Hz";
  std::string str1 = ou1.str();

  std::ostringstream ou2;
  ou2 << "NF-" << ((int)mBandFreqB) << "-" << ((int)mBandFreqA) << "Hz";
  std::string str2 = ou2.str();

  std::ostringstream ou3;
  ou3 << "NF-" << ((int)mBandFreqC) << "-" << ((int)mBandFreqB) << "Hz";
  std::string str3 = ou3.str();

  std::ostringstream ou4;
  ou4 << "NF-" << ((int)mBandFreqD) << "-" << ((int)mBandFreqC) << "Hz";
  std::string str4 = ou4.str();

  std::ostringstream ou5;
  ou5 << "NF-0-" << ((int)mBandFreqA) << "-Hz-12hs";
  std::string str5 = ou5.str();

  std::ostringstream ou6;
  ou6 << "NF-" << ((int)mBandFreqB) << "-" << ((int)mBandFreqA) << "Hz-12hs";
  std::string str6 = ou6.str();

  std::ostringstream ou7;
  ou7 << "NF-" << ((int)mBandFreqC) << "-" << ((int)mBandFreqB) << "Hz-12hs";
  std::string str7 = ou7.str();

  std::ostringstream ou8;
  ou8 << "NF-" << ((int)mBandFreqD) << "-" << ((int)mBandFreqC) << "Hz-12hs";
  std::string str8 = ou8.str();


  //----------------------------------  Specify DMT Viewer channel names
  //--------------------------------------------------------------------
  std::ostringstream chNout;
  chNout << "NoiseFloor-" << mDetectorChannel;
  std::string chN = chNout.str();
  setServerName(chN.c_str());
  MonServer::serveData(str1.c_str(), &mNoiseFloorBandA);
  MonServer::serveData(str2.c_str(), &mNoiseFloorBandB);
  MonServer::serveData(str3.c_str(), &mNoiseFloorBandC);
  MonServer::serveData(str4.c_str(), &mNoiseFloorBandD);
  MonServer::serveData(str5.c_str(), &mNoiseFloorBandA12);
  MonServer::serveData(str6.c_str(), &mNoiseFloorBandB12);
  MonServer::serveData(str7.c_str(), &mNoiseFloorBandC12);
  MonServer::serveData(str8.c_str(), &mNoiseFloorBandD12);
  MonServer::serveData("LowerThres-BandA", &mLowerThresholdBandA);
  MonServer::serveData("UpperThres-BandA", &mUpperThresholdBandA);
  MonServer::serveData("LowerThres-BandB", &mLowerThresholdBandB);
  MonServer::serveData("UpperThres-BandB", &mUpperThresholdBandB);
  MonServer::serveData("LowerThres-BandC", &mLowerThresholdBandC);
  MonServer::serveData("UpperThres-BandC", &mUpperThresholdBandC);
  MonServer::serveData("LowerThres-BandD", &mLowerThresholdBandD);
  MonServer::serveData("UpperThres-BandD", &mUpperThresholdBandD);
  MonServer::serveData("Lower2Sigma-BandA", &mLower2SigmaBandA);
  MonServer::serveData("Upper2Sigma-BandA", &mUpper2SigmaBandA);
  MonServer::serveData("Lower2Sigma-BandB", &mLower2SigmaBandB);
  MonServer::serveData("Upper2Sigma-BandB", &mUpper2SigmaBandB);
  MonServer::serveData("Lower2Sigma-BandC", &mLower2SigmaBandC);
  MonServer::serveData("Upper2Sigma-BandC", &mUpper2SigmaBandC);
  MonServer::serveData("Lower2Sigma-BandD", &mLower2SigmaBandD);
  MonServer::serveData("Upper2Sigma-BandD", &mUpper2SigmaBandD);

  //----------------------------------  Set up the trender
  //--------------------------------------------------------------------
  std::string IFO(&mDetectorChannel[0],2);
  mTrend.setIFO(IFO.c_str());
  std::ostringstream trN;
  trN << "NoiseFloor_" << IFO;
  mTrend.setName(trN.str().c_str());    //  Set trend frame name
  mTrend.setFrameCount(1);          //  Set frames per file
  mTrend.setType(Trend::kMinute);   //  Set trend type (minute trends)
  mTrend.setAutoUpdate(false);      //  Avoid automatic screwup.

  //----------------------------------  Specify trend channels
  //----------------------------------  Band A
  //--------------------------------------------------------------------
  std::ostringstream trNp;
  std::ostringstream trNm;
  trNp << IFO << ":DMT-NFLM_" << 0 << "_" << mBandFreqA << "_PERCENTAGE";
  mPercentageChannelBandA = trNp.str();
  //  std::cout << " ... trend channel " << mPercentageChannelBandA << std::endl;
  mTrend.addChannel(mPercentageChannelBandA.c_str());
  trNm << IFO << ":DMT-NFLM_" << 0 << "_" << mBandFreqA << "_MAXVALUE";
  mMaxValueChannelBandA = trNm.str();
  //  std::cout << " ... trend channel " << mMaxValueChannelBandA << std::endl;
  mTrend.addChannel(mMaxValueChannelBandA.c_str());
  //----------------------------------  Band B
  //--------------------------------------------------------------------
  trNp.seekp(0);
  trNp << IFO << ":DMT-NFLM_" << mBandFreqA << "_" << mBandFreqB << "_PERCENTAGE";
  mPercentageChannelBandB = trNp.str();
  //  std::cout << " ... trend channel " << mPercentageChannelBandB << std::endl;
  mTrend.addChannel(mPercentageChannelBandB.c_str());
  trNm.seekp(0);
  trNm << IFO << ":DMT-NFLM_" << mBandFreqA << "_" << mBandFreqB << "_MAXVALUE";
  mMaxValueChannelBandB = trNm.str();
  //  std::cout << " ... trend channel " << mMaxValueChannelBandB << std::endl;
  mTrend.addChannel(mMaxValueChannelBandB.c_str());
  //----------------------------------  Band C
  //--------------------------------------------------------------------
  trNp.seekp(0);
  trNp << IFO << ":DMT-NFLM_" << mBandFreqB << "_" << mBandFreqC << "_PERCENTAGE";
  mPercentageChannelBandC = trNp.str();
  //  std::cout << " ... trend channel " << mPercentageChannelBandC << std::endl;
  mTrend.addChannel(mPercentageChannelBandC.c_str());
  trNm.seekp(0);
  trNm << IFO << ":DMT-NFLM_" << mBandFreqB << "_" << mBandFreqC << "_MAXVALUE";
  mMaxValueChannelBandC = trNm.str();
  //  std::cout << " ... trend channel " << mMaxValueChannelBandC << std::endl;
  mTrend.addChannel(mMaxValueChannelBandC.c_str());
  //----------------------------------  Band D
  //--------------------------------------------------------------------
  trNp.seekp(0);
  trNp << IFO << ":DMT-NFLM_" << mBandFreqC << "_" << mBandFreqD << "_PERCENTAGE";
  mPercentageChannelBandD = trNp.str();
  //  std::cout << " ... trend channel " << mPercentageChannelBandD << std::endl;
  mTrend.addChannel(mPercentageChannelBandD.c_str());
  trNm.seekp(0);
  trNm << IFO << ":DMT-NFLM_" << mBandFreqC << "_" << mBandFreqD << "_MAXVALUE";
  mMaxValueChannelBandD = trNm.str();
  //  std::cout << " ... trend channel " << mMaxValueChannelBandD << std::endl;
  mTrend.addChannel(mMaxValueChannelBandD.c_str());

  //----------------------------------  Calculat plot size for the twelve hs plot
  //-----------------------------------------------------------------------------
  unsigned int size12h = (unsigned int)(43200./mTStride.GetSecs());
  for (unsigned int ii = 0; ii < size12h; ii++)
  {
    m12bandA.push_back(double(0));
    m12bandB.push_back(double(0));
    m12bandC.push_back(double(0));
    m12bandD.push_back(double(0));
  }

  //----------------------------------  Set Operation State Condition
  //--------------------------------------------------------------------
  mOSC = new OperStateCondList(getDacc());
  mOSC->setStride((int)mTStride.GetSecs());
  mOSC->setDebug(0);
  mOSC->ignoreAllExcept(mDetectorChannel.c_str());
  mOSC->readConfig("LockLoss.conf");
  mLockConditionFlag = true;

  //----------------------------------  Set file names for html output
  //--------------------------------------------------------------------
  const char* thepath  = getenv("DMTHTMLOUT");
  std::string htmlPath;
  if (!thepath)
  {
    std::cerr << "ERROR: can't get html directory from environment variable DMTHTMLOUT" << std::endl;
    exit(1);
  }
  else
    htmlPath = std::string(thepath);
  // Check if environment variable is defined
  if (htmlPath.empty())
  {
    std::cerr << "ERROR: can't set html directory from environment variable DMTHTMLOUT" << std::endl;
    exit(1);
  }
  if (htmlPath.at(htmlPath.size()-1) != '/')
    htmlPath.append("/");

  // check if directory exists and file can open and written
  //mHTMLStatusFile = htmlPath + mDetectorChannel.substr(0,2) + std::string("NFTMStatusPage.html");
  mHTMLStatusFile = htmlPath + std::string("NFTMStatusPage.html");
  mHTMLErrorFile  = htmlPath + std::string("NFTMErrorPage.html");
  std::ofstream outStatusF(mHTMLStatusFile.c_str());
  std::ofstream outErrorF(mHTMLErrorFile.c_str());
  if (!outStatusF || !outErrorF)
  {
    std::cerr << "ERROR: can't open html file for monitor status and error report!" << std::endl;
    exit(1);
  }
  else
  {
    outStatusF.close();
    outErrorF.close();
  }
} // Constructor

//======================================
//======================================  Skeleton object destructor.
NoiseFloorMonitor::~NoiseFloorMonitor()
{
  std::cout << "NoiseFloorMonitor is finished" << std::endl;
}

//======================================
//======================================  Handle Messages
void
NoiseFloorMonitor::Attention(void)
{
    MonServer::Attention();
}

//======================================
//======================================  Frame processing function.
void
NoiseFloorMonitor::ProcessData(void)
{
  // std::cout << std::endl << std::endl;
  // msr::Clock clk;
  // clk.start();
  
  //----------------------------------  clear error buffer
  //-------------------------------------------------------------------------------
  gwd::Singleton* single = gwd::Singleton::exemplar();
  single->ClearMessages();

  //----------------------------------  Get channel data: time series
  //--------------------------------------------------------------------
  bool tsFlag = true;
  std::vector<double> inTS;
  //----------------------------------  Get pointers to the current data
  //--------------------------------------------------------------------
  const TSeries* ts = getDacc().refData(mDetectorChannel.c_str());
  //----------------------------------  Check if data is correct
  //--------------------------------------------------------------------
  if (ts == 0)
  {
    tsFlag = false;
    single->AppendMessage("NoiseFloorMonitor::ProcessData(): time series pointer is NULL");
  }
  else if (ts->isEmpty())
  {
    tsFlag = false;
    single->AppendMessage("NoiseFloorMonitor::ProcessData(): time series is empty");
  }
  else if (std::string(ts->getName()) != mDetectorChannel)
  {
    tsFlag = false;
    single->AppendMessage(std::string("NoiseFloorMonitor::ProcessData(): get data from wrong channel = ")+std::string(ts->getName()));
  }

  if (tsFlag)
  {
    //----------------------------------  Get current time
    //--------------------------------------------------------------------
    mCurrentTime = ts->getStartTime();

    //----------------------------------  remember no. of samples
    //--------------------------------------------------------------------
    const unsigned int tsNSample = ts->getNSample();

    //----------------------------------  check sample rate of input ts
    //--------------------------------------------------------------------
    double sampleRate = 1./ ts->getTStep().GetSecs();
    if (fabs(sampleRate-mSignalSampleRate) > std::numeric_limits<float>::epsilon())
    {
      //----------------------------------  update signal sampling rate
      //--------------------------------------------------------------------
      mSignalSampleRate = sampleRate;

      // needs to re-compute all filters
      //----------------------------------  compute the cutoff filter
      //--------------------------------------------------------------------
      gwd::LowPassFilter  lpf;
      lpf.ComputeFilter(mSignalSampleRate,mUpperCutoffFreq,mFltOrder,mCutoffFlt);

      //----------------------------------  compute Notch filter
      //--------------------------------------------------------------------
      gwd::NotchFilter nf;
      nf.ComputeFilter(mSampleRate,mOrderNotchFilter,mFreqNotchFilter,mFreqBndWidthNotFilter,mNotchFilter);
      
      //----------------------------------  write warning message
      //--------------------------------------------------------------------
      single->AppendMessage("NoiseFloorMonitor::ProcessData(): sample rate has changed, cutoff and Notch filter were re-computed");
    }

    //----------------------------------  set buffer
    //--------------------------------------------------------------------
    inTS.resize(buffer.size()+tsNSample);
    std::copy(buffer.begin(),buffer.end(),inTS.begin());
    std::vector<double>::iterator p = inTS.begin()+buffer.size();
    for (unsigned int ii = 0; ii < tsNSample; ii++) *p++ = ts->getDouble(ii);
    std::copy(inTS.end()-buffer.size(),inTS.end(),buffer.begin());

    // check if time series is empty
    if (inTS.empty())
    {
      tsFlag = false;
      single->AppendMessage("ProcessData(): can't create input time series from detector data");
    }
  } // if (tsFlag)

  //----------------------------------  Check detector's operation state condition
  //------------------------------------------------------------------------------
  if (mOSC->satisfied(mLockConditionName.c_str()))
  {
    mLockConditionFlag = true;
  }
  else
  {
    mLockConditionFlag = false;
  }

  // Main Routine: compute noise floor
  if (tsFlag && mLockConditionFlag)
  {
    bool flag = true;
    //----------------------------------  Resample and FIR filtering
    //--------------------------------------------------------------------
    gwd::FIRFilter fir;
    gwd::Multirate mrt;

    //----------------------------------  compute band limited data:
    //----------------------------------  16kHz -> 4kHz sample rate
    //--------------------------------------------------------------------
    std::vector<double> locTS;
    int factor = (int)gwd::NearestInt(mSignalSampleRate/mSampleRate);
    flag = mrt.Decimate(factor,mCutoffFlt,inTS,locTS);
   
    //----------------------------------  compute the whitening filter
    //--------------------------------------------------------------------
    // Sun does not like unsigned int, have to use ints!!!!
    unsigned int a = (unsigned int) mCounterWitheFilter;
    unsigned int b = (unsigned int) mUpdateWhiteFlt;
    if ((a%b) == 0)
    {
      gwd::WhiteningFilter wf;
      const unsigned int wSize    = (unsigned int) (mRMSizeForWhiteningFlt);
      const unsigned int segSize  = (unsigned int) (mSampleRate * mSizePsd) + wSize;
      const unsigned int tshift   = (unsigned int) (5. * mSampleRate)+ buffer.size(); // shift by 5 secs. + buffer size

      //----------------------------------  create a vector with the time series
      //--------------------------------------------------------------------
      std::vector<double> locSig(segSize);
      for (unsigned int nn = 0; nn < segSize; nn++)
        locSig[nn] = locTS[tshift+nn];

      //----------------------------------  Compute the whiten filter
      //--------------------------------------------------------------------
      if (!wf.ComputeFilter(mOrderWhiteFilter,mNoOfFreqPsd,wSize,locSig,mWhiteningFilter))
      {
        single->AppendMessage("NoiseFloorMonitor::ProcessData(): can't compute whitening filter");
        flag = false;
        // tray the next loop
        mCounterWitheFilter = 0;
      }
    }

    //----------------------------------  Whitening
    //--------------------------------------------------------------------
    std::vector<double> wtTS;
    if (flag) flag = fir.Filter(locTS,mWhiteningFilter,wtTS);
    if (!flag)
    {
      single->AppendMessage("NoiseFloorMonitor::ProcessData(): can't whiten data");
      flag = false;
    }

    //----------------------------------  line removal
    //--------------------------------------------------------------------
    std::vector<double> lrTS; // line removed signal
    if (flag) flag = fir.Filter(wtTS,mNotchFilter,lrTS);
    if (!flag)
    {
      single->AppendMessage("NoiseFloorMonitor::ProcessData(): can't eliminate lines with Notch filter");
      flag = false;
    }

    //----------------------------------  computes four bands
    //--------------------------------------------------------------------
    std::vector<double> bandB;
    std::vector<double> bandC;
    std::vector<double> bandD;
    std::vector<double> ss;

    //----------------------------------  process band A
    //--------------------------------------------------------------------
    //  std::cout << " ... processing frequency band A" << std::endl;
    std::vector<double> bandA;
    mrt.Decimate(mDecimateFactorA,mLpFltA,lrTS,bandA);

    //----------------------------------  process band B
    //--------------------------------------------------------------------
    //  std::cout << " ... processing frequency band B" << std::endl;
    double bandwidth = mBandFreqB - mBandFreqA;
    fir.Filter(lrTS,mBpFltB,ss);                               // band-pass filter
    MixWithOscilator(bandwidth/2.,mSampleRate,ss);              // mix with oscillator
    mrt.Decimate(mDecimateFactorB,mLpFltB,ss,bandB);            // decimate
    ss.clear();
    //----------------------------------  process band C
    //--------------------------------------------------------------------
    //  std::cout << " ... processing frequency band C" << std::endl;
    bandwidth = mBandFreqC - mBandFreqB;
    fir.Filter(lrTS,mBpFltC,ss);                               // band-pass filter
    MixWithOscilator(bandwidth/2.,mSampleRate,ss);              // mix with oscillator
    mrt.Decimate(mDecimateFactorC,mLpFltC,ss,bandC);            // decimate
    ss.clear();
    //----------------------------------  process band D
    //--------------------------------------------------------------------
    //  std::cout << " ... processing frequency band D" << std::endl;
    bandwidth = mBandFreqD - mBandFreqC;
    fir.Filter(lrTS,mHpFltD,ss);                               // high-pass filter
    MixWithOscilator(bandwidth/2.,mSampleRate,ss);              // mix with oscillator
    mrt.Decimate(mDecimateFactorD,mLpFltD,ss,bandD);            // decimate
    ss.clear();

    //----------------------------------  select the time stride
    //--------------------------------------------------------------------
    std::vector<double> tstride;
    unsigned int tsShift;
    unsigned int tsSize;
    const double szTStride = mTStride.GetSecs();
    //----------------------------------  time stride band A
    //--------------------------------------------------------------------
    tsShift = static_cast<unsigned int>((mFilterDelay+mBufferTimeShift)*mSampleRateBandA);
    tsSize  = static_cast<unsigned int>((szTStride+mWindowSizeRunningMedian)*mSampleRateBandA);
    tstride.resize(tsSize);
    for (unsigned int ii = 0; ii < tsSize; ii++) tstride[ii] = bandA[tsShift+ii];
    bandA.resize(tsSize);
    std::copy(tstride.begin(),tstride.end(),bandA.begin());
    tstride.clear();

    //----------------------------------  time stride band B
    //--------------------------------------------------------------------
    tsShift = static_cast<unsigned int>((mFilterDelay+mBufferTimeShift)*mSampleRateBandB);
    tsSize  = static_cast<unsigned int>((szTStride+mWindowSizeRunningMedian)*mSampleRateBandB);
    tstride.resize(tsSize);
    for (unsigned int ii = 0; ii < tsSize; ii++) tstride[ii] = bandB[tsShift+ii];
    bandB.resize(tsSize);
    std::copy(tstride.begin(),tstride.end(),bandB.begin());
    tstride.clear();
    
    //----------------------------------  time stride band C
    //--------------------------------------------------------------------
    tsShift = static_cast<unsigned int>((mFilterDelay+mBufferTimeShift)*mSampleRateBandC);
    tsSize  = static_cast<unsigned int>((szTStride+mWindowSizeRunningMedian)*mSampleRateBandC);
    tstride.resize(tsSize);
    for (unsigned int ii = 0; ii < tsSize; ii++) tstride[ii] = bandC[tsShift+ii];
    bandC.resize(tsSize);
    std::copy(tstride.begin(),tstride.end(),bandC.begin());
    tstride.clear();
    
    //----------------------------------  time stride band D
    //--------------------------------------------------------------------
    tsShift = static_cast<unsigned int>((mFilterDelay+mBufferTimeShift)*mSampleRateBandD);
    tsSize  = static_cast<unsigned int>((szTStride+mWindowSizeRunningMedian)*mSampleRateBandD);
    tstride.resize(tsSize);
    for (unsigned int ii = 0; ii < tsSize; ii++) tstride[ii] = bandD[tsShift+ii];
    bandD.resize(tsSize);
    std::copy(tstride.begin(),tstride.end(),bandD.begin());
    tstride.clear();

    //----------------------------------  transform to Normal(0,1)
    //--------------------------------------------------------------------
    double mean;
    double stdDev;
    Estimator(bandA,mean,stdDev);
    for (size_t nn = 0; nn < bandA.size(); nn++) bandA[nn] = (bandA[nn] - mean)/stdDev;

    Estimator(bandB,mean,stdDev);
    for (size_t nn = 0; nn < bandB.size(); nn++) bandB[nn] = (bandB[nn] - mean)/stdDev;

    Estimator(bandC,mean,stdDev);
    for (size_t nn = 0; nn < bandC.size(); nn++) bandC[nn] = (bandC[nn] - mean)/stdDev;

    Estimator(bandD,mean,stdDev);
    for (size_t nn = 0; nn < bandD.size(); nn++) bandD[nn] = (bandD[nn] - mean)/stdDev;

    //----------------------------------  square the values
    //--------------------------------------------------------------------
    for (size_t nn =0; nn < bandA.size(); nn++) bandA[nn] = gwd::square(bandA[nn]);
    for (size_t nn =0; nn < bandB.size(); nn++) bandB[nn] = gwd::square(bandB[nn]);
    for (size_t nn =0; nn < bandC.size(); nn++) bandC[nn] = gwd::square(bandC[nn]);
    for (size_t nn =0; nn < bandD.size(); nn++) bandD[nn] = gwd::square(bandD[nn]);

    //----------------------------------  running median
    //--------------------------------------------------------------------
    //  std::cout << " ... computing running median" << std::endl;
    gwd::RunningMedian rm;
    unsigned int rmwz;
    std::vector<double> rmBandA;
    rmwz = (unsigned int)(mWindowSizeRunningMedian*mSampleRateBandA);
    rm.ComputeMedians(rmwz,bandA,rmBandA);

    std::vector<double> rmBandB;
    rmwz = (unsigned int)(mWindowSizeRunningMedian*mSampleRateBandB);
    rm.ComputeMedians(rmwz,bandB,rmBandB);

    std::vector<double> rmBandC;
    rmwz = (unsigned int)(mWindowSizeRunningMedian*mSampleRateBandC);
    rm.ComputeMedians(rmwz,bandC,rmBandC);

    std::vector<double> rmBandD;
    rmwz = (unsigned int)(mWindowSizeRunningMedian*mSampleRateBandD);
    rm.ComputeMedians(rmwz,bandD,rmBandD);

    //---------------------------------- shift data to zero mean
    //--------------------------------------------------------------------
    Estimator(rmBandA,mean,stdDev);
    for (size_t nn = 0; nn < rmBandA.size(); nn++) rmBandA[nn] = (rmBandA[nn] - mean);

    Estimator(rmBandB,mean,stdDev);
    for (size_t nn = 0; nn < rmBandB.size(); nn++) rmBandB[nn] = (rmBandB[nn] - mean);

    Estimator(rmBandC,mean,stdDev);
    for (size_t nn = 0; nn < rmBandC.size(); nn++) rmBandC[nn] = (rmBandC[nn] - mean);

    Estimator(rmBandD,mean,stdDev);
    for (size_t nn = 0; nn < rmBandD.size(); nn++) rmBandD[nn] = (rmBandD[nn] - mean);

    //----------------------------------  send data to dmt-viewer
    //--------------------------------------------------------------------
    TimeSeriesPlot(rmBandA,rmBandB,rmBandC,rmBandD);

    //----------------------------------  set trend and twelve hours plot
    //--------------------------------------------------------------------
    SetNoiseFloorTrend(rmBandA,rmBandB,rmBandC,rmBandD);

    //----------------------------------  Count locked loops for update of whiten filter
    //----------------------------------------------------------------------------------
    mCounterWitheFilter++;
  } // LockConditionFlag
  else
  {
    //----------------------------------  Send empty plot to the dmt-viewer
    //----------------------------------------------------------------------------------
    EmptyPlot();
  } // LockConditionFlag

  //----------------------------------  Plot threshold values
  //----------------------------------------------------------------------------------
  ThresholdPlot();

  //----------------------------------  output processing time
  //--------------------------------------------------------------------
  // std::ostringstream ouString;
  // ouString << " ... Loop: " << ((int)mCounterTimeStrides);
  // std::string str = ouString.str();

  //----------------------------------  get processing time
  //--------------------------------------------------------------------
  // clk.stop();
  // clk.print(str.c_str());

  //----------------------------------  Write Monitor Status in html
  //--------------------------------------------------------------------
  DumpHTMLStatusPage();

  //----------------------------------  get processing time
  //--------------------------------------------------------------------
  DumpHTMLErrorPage();

  //----------------------------------  update counter
  //--------------------------------------------------------------------
  mCounterTimeStrides++;
}


//======================================
//====================================== Estimate Threshold from Monte Carlo simulation
void
NoiseFloorMonitor::EstimateThreshold( )
{
  // Calculate threshold for band A
  //========================================================================================
  unsigned int rmwz = static_cast<unsigned int>(mWindowSizeRunningMedian*mSampleRateBandA);
  unsigned int lcSz = static_cast<unsigned int>(mTStride.GetSecs()*mSampleRateBandA);

  std::vector<double> noise(lcSz);
  gwd::NormalDistribution normal;

  for(unsigned int ii = 0; ii < lcSz; ii++)
  {
    noise[ii] = normal.Normal(0.,1.);
    noise[ii] = gwd::square(noise[ii]);
  }
  gwd::RunningMedian rm;
  std::vector<double> rmTS;
  rm.ComputeMedians(rmwz,noise,rmTS);

  // Compute mean and std. deviation
  Estimator(rmTS,mMeanA,mStdDevA);

  // Calculate threshold for band B
  //========================================================================================
  rmwz = static_cast<unsigned int>(mWindowSizeRunningMedian*mSampleRateBandB);
  lcSz = static_cast<unsigned int>(mTStride.GetSecs()*mSampleRateBandB);

  noise.resize(lcSz);
  for(unsigned int ii = 0; ii < lcSz; ii++)
  {
    noise[ii] = normal.Normal(0.,1.);
    noise[ii] = gwd::square(noise[ii]);
  }

  rmTS.clear();
  rm.ComputeMedians(rmwz,noise,rmTS);

  // Compute mean and std. deviation
  Estimator(rmTS,mMeanB,mStdDevB);

  // Calculate threshold for band C
  //========================================================================================
  rmwz = static_cast<unsigned int>(mWindowSizeRunningMedian*mSampleRateBandC);
  lcSz = static_cast<unsigned int>(mTStride.GetSecs()*mSampleRateBandC);

  noise.resize(lcSz);
  for(unsigned int ii = 0; ii < lcSz; ii++)
  {
    noise[ii] = normal.Normal(0.,1.);
    noise[ii] = gwd::square(noise[ii]);
  }

  rmTS.clear();
  rm.ComputeMedians(rmwz,noise,rmTS);

  // Compute mean and std. deviation
  Estimator(rmTS,mMeanC,mStdDevC);


  // Calculate threshold for band D
  //========================================================================================
  rmwz = static_cast<unsigned int>(mWindowSizeRunningMedian*mSampleRateBandD);
  lcSz = static_cast<unsigned int>(mTStride.GetSecs()*mSampleRateBandD);

  noise.resize(lcSz);
  for(unsigned int ii = 0; ii < lcSz; ii++)
  {
    noise[ii] = normal.Normal(0.,1.);
    noise[ii] = gwd::square(noise[ii]);
  }

  rmTS.clear();
  rm.ComputeMedians(rmwz,noise,rmTS);

  // Compute mean and std. deviation
  Estimator(rmTS,mMeanD,mStdDevD);
}


//======================================
//====================================== Prepare data for plot
void
NoiseFloorMonitor::PrepareDataForPlot(std::vector<double>& in,std::vector<float>& out)
{
  //
  // The dmtviewer can't plot more than 10000 samples at a time
  //###################################################################
  const unsigned int maxNoOfSamples = 10000;

  // obtain the time stride to be plotted
  unsigned int inSize;
  unsigned int ouSize;

  // Check no. of samples
  inSize = (unsigned int) in.size();
  ouSize = inSize;
  // Distinguish two cases:
  //  1. input signal is larger than output signal
  //     here resampling is necessary
  //  2. the input signal is shortar than maxNoOfSamples
  //     simply copy input into output
  if (ouSize < maxNoOfSamples)
  {
    out.resize(ouSize);
    for (unsigned int ii = 0; ii < ouSize; ii++) out[ii] = in[ii];
    return;
  }
  else
  {
    // can't write more samples than maxNoOfSamples
    ouSize = maxNoOfSamples;
    // Time intervals
    const double dt = mTStride / static_cast<double>(inSize - 1);
    const double DT = mTStride / static_cast<double>(ouSize);
    // get min-max
    double inMax = *(std::max_element(in.begin(),in.end()));
    double inMin = *(std::min_element(in.begin(),in.end()));
    
    // Check if signal has a constant value
    if (fabs(inMin - inMax) <= std::numeric_limits<float>::epsilon( ))
    {
      out.resize(ouSize);
      for (unsigned int ii = 1; ii < ouSize; ii++) out[ii] = in[ii];
    }
    else
    {
      // Set output data
      out.resize(ouSize);
      out[0] = in[0];
      for (unsigned int ii = 1; ii < ouSize; ii++)
      {
        // Find time in [0,mTStride]
        double t = static_cast<double>(ii)*DT;
        unsigned int jj = (unsigned int)floor(t/dt);
        double delta = t - static_cast<double>(jj)*dt;
        // Linear interpolate value
        out[ii] = static_cast<float>(((in[jj+1]-in[jj]) / dt)*delta + in[jj]);
      }
  
      // Scale the signal to min-max interval
      float outMax = *(std::max_element(out.begin(),out.end()));
      float outMin = *(std::min_element(out.begin(),out.end()));
      float alpha;
      float beta;
      if (fabs(outMin - outMax) <= std::numeric_limits<float>::epsilon( ))
      {
        alpha = 1.;
        beta  = 0.;
      }
      else
      {
        alpha = (outMax - outMin) / (inMax - inMin);
        beta  = (inMax*outMin - inMin*outMax) / (inMax - inMin);
      }
      
      for (unsigned int ii = 0; ii < ouSize; ii++)
      {
        out[ii] = alpha*out[ii] + beta;
      }
    }
    // finish!
    return;
  }
}

//======================================
//====================================== DMT-viewer: Empty plot for non locked state
void 
NoiseFloorMonitor::EmptyPlot()
{
  //----------------------------------  Plot empty time series for non locked state
  //----------------------------------------------------------------------------------
  Time t0 = mCurrentTime;
  Interval dt;
  std::vector<float> ouTS(1000);
  dt = mTStride / static_cast<double>(ouTS.size()-1);
  std::fill(ouTS.begin(),ouTS.end(),double(0));
  mNoiseFloorBandA.setData(t0, dt, &ouTS[0], ouTS.size());
  mNoiseFloorBandB.setData(t0, dt, &ouTS[0], ouTS.size());
  mNoiseFloorBandC.setData(t0, dt, &ouTS[0], ouTS.size());
  mNoiseFloorBandD.setData(t0, dt, &ouTS[0], ouTS.size());

  //----------------------------------  Process twelve hours plots
  //----------------------------------------------------------------------------------
  // For the twelve hour plot use 12hs = 43200 sec
  Time t12 = mCurrentTime - Interval(43200.);
  Interval dt12;
  //----------------------------------  12hs Band A
  m12bandA.pop_front();
  m12bandA.push_back(double(0));
  ouTS.resize(m12bandA.size());
  std::copy(m12bandA.begin(),m12bandA.end(),ouTS.begin());
  dt12 = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandA12.setData(t12, dt12, &ouTS[0], ouTS.size());
  //----------------------------------  12hs Band B
  m12bandB.pop_front();
  m12bandB.push_back(double(0));
  ouTS.resize(m12bandB.size());
  std::copy(m12bandB.begin(),m12bandB.end(),ouTS.begin());
  dt12 = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandB12.setData(t12, dt12, &ouTS[0], ouTS.size());
  //----------------------------------  12hs Band C
  m12bandC.pop_front();
  m12bandC.push_back(double(0));
  ouTS.resize(m12bandC.size());
  std::copy(m12bandC.begin(),m12bandC.end(),ouTS.begin());
  dt12 = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandC12.setData(t12, dt12, &ouTS[0], ouTS.size());
  //----------------------------------  12hs Band D
  m12bandD.pop_front();
  m12bandD.push_back(double(0));
  ouTS.resize(m12bandD.size());
  std::copy(m12bandD.begin(),m12bandD.end(),ouTS.begin());
  dt12 = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandD12.setData(t12, dt12, &ouTS[0], ouTS.size());
  
  //----------------------------------  write zero trend data
  //----------------------------------------------------------------------------------
  //----------------------------------  Band A
  mTrend.trendData(mPercentageChannelBandA.c_str(),t0,double(0));
  mTrend.trendData(mMaxValueChannelBandA.c_str(),t0,double(0));
  //----------------------------------  Band B
  mTrend.trendData(mPercentageChannelBandB.c_str(),t0,double(0));
  mTrend.trendData(mMaxValueChannelBandB.c_str(),t0,double(0));
  //----------------------------------  Band C
  mTrend.trendData(mPercentageChannelBandC.c_str(),t0,double(0));
  mTrend.trendData(mMaxValueChannelBandC.c_str(),t0,double(0));
  //----------------------------------  Band D
  mTrend.trendData(mPercentageChannelBandD.c_str(),t0,double(0));
  mTrend.trendData(mMaxValueChannelBandD.c_str(),t0,double(0));
}

//======================================
//====================================== DMT-viewer: threshold plot
void 
NoiseFloorMonitor::ThresholdPlot()
{
  Time t0 = mCurrentTime;
  std::vector<float> ouTS(1000);
  Interval dt = mTStride / static_cast<double>(ouTS.size()-1);
  //----------------------------------  Band A
  std::fill(ouTS.begin(),ouTS.end(),3.*mStdDevA);
  mUpperThresholdBandA.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-3.*mStdDevA);
  mLowerThresholdBandA.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),2.*mStdDevA);
  mUpper2SigmaBandA.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-2.*mStdDevA);
  mLower2SigmaBandA.setData(t0, dt, &ouTS[0], ouTS.size());
  //----------------------------------  Band B
  std::fill(ouTS.begin(),ouTS.end(),3.*mStdDevB);
  mUpperThresholdBandB.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-3.*mStdDevB);
  mLowerThresholdBandB.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),2.*mStdDevB);
  mUpper2SigmaBandB.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-2.*mStdDevB);
  mLower2SigmaBandB.setData(t0, dt, &ouTS[0], ouTS.size());
  //----------------------------------  Band C
  std::fill(ouTS.begin(),ouTS.end(),3.*mStdDevC);
  mUpperThresholdBandC.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-3.*mStdDevC);
  mLowerThresholdBandC.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),2.*mStdDevC);
  mUpper2SigmaBandC.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-2.*mStdDevC);
  mLower2SigmaBandC.setData(t0, dt, &ouTS[0], ouTS.size());
  //----------------------------------  Band D
  std::fill(ouTS.begin(),ouTS.end(),3.*mStdDevD);
  mUpperThresholdBandD.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-3.*mStdDevD);
  mLowerThresholdBandD.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),2.*mStdDevD);
  mUpper2SigmaBandD.setData(t0, dt, &ouTS[0], ouTS.size());
  std::fill(ouTS.begin(),ouTS.end(),-2.*mStdDevD);
  mLower2SigmaBandD.setData(t0, dt, &ouTS[0], ouTS.size());
}

//======================================
//====================================== DMT-viewer: ime series plot
void 
NoiseFloorMonitor::TimeSeriesPlot(std::vector<double>& rmBandA,std::vector<double>& rmBandB,
                                  std::vector<double>& rmBandC,std::vector<double>& rmBandD)
{
  //----------------------------------  send data to dmt-viewer
  //--------------------------------------------------------------------
  Time t0 = mCurrentTime;
  Interval dt;
  std::vector<float> ouTS;
  PrepareDataForPlot(rmBandA,ouTS);
  dt = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandA.setData(t0, dt, &ouTS[0], ouTS.size());

  PrepareDataForPlot(rmBandB,ouTS);
  dt = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandB.setData(t0, dt, &ouTS[0], ouTS.size());

  PrepareDataForPlot(rmBandC,ouTS);
  dt = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandC.setData(t0, dt, &ouTS[0], ouTS.size());

  PrepareDataForPlot(rmBandD,ouTS);
  dt = mTStride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandD.setData(t0, dt, &ouTS[0], ouTS.size());
}

//======================================
//====================================== Set the noise floor trend and 12hs plot
void 
NoiseFloorMonitor::SetNoiseFloorTrend(std::vector<double>& rmBandA,std::vector<double>& rmBandB,
                                      std::vector<double>& rmBandC,std::vector<double>& rmBandD)  
{
  // For the twelve hour plot use always
  // 12hs = 43200 sec
  Time t0 = mCurrentTime;
  Time t12 = mCurrentTime - Interval(43200.);
  Interval dt12;
  Interval stride(double(43200));
  // data buffer
  std::vector<double> ouTS;
  
  //----------------------------------  Trend the data points
  //--------------------------------------------------------------------
  double maxVal,percentage;
  //----------------------------------  Band A
  GetTrend(mStdDevA,rmBandA,maxVal,percentage);
  mTrend.trendData(mPercentageChannelBandA.c_str(),t0,percentage);
  mTrend.trendData(mMaxValueChannelBandA.c_str(),t0,maxVal);
  m12bandA.pop_front();
  m12bandA.push_back(percentage);
  ouTS.resize(m12bandA.size());
  std::copy(m12bandA.begin(),m12bandA.end(),ouTS.begin());
  dt12 = stride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandA12.setData(t12, dt12, &ouTS[0], ouTS.size());

  //----------------------------------  Band B
  GetTrend(mStdDevB,rmBandB,maxVal,percentage);
  mTrend.trendData(mPercentageChannelBandB.c_str(),t0,percentage);
  mTrend.trendData(mMaxValueChannelBandB.c_str(),t0,maxVal);
  m12bandB.pop_front();
  m12bandB.push_back(percentage);
  ouTS.resize(m12bandB.size());
  std::copy(m12bandB.begin(),m12bandB.end(),ouTS.begin());
  dt12 = stride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandB12.setData(t12, dt12, &ouTS[0], ouTS.size());

  //----------------------------------  Band C
  GetTrend(mStdDevC,rmBandC,maxVal,percentage);
  mTrend.trendData(mPercentageChannelBandC.c_str(),t0,percentage);
  mTrend.trendData(mMaxValueChannelBandC.c_str(),t0,maxVal);
  m12bandC.pop_front();
  m12bandC.push_back(percentage);
  ouTS.resize(m12bandC.size());
  std::copy(m12bandC.begin(),m12bandC.end(),ouTS.begin());
  dt12 = stride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandC12.setData(t12, dt12, &ouTS[0], ouTS.size());

  //----------------------------------  Band D
  GetTrend(mStdDevD,rmBandD,maxVal,percentage);
  mTrend.trendData(mPercentageChannelBandD.c_str(),t0,percentage);
  mTrend.trendData(mMaxValueChannelBandD.c_str(),t0,maxVal);
  mTrend.Update();
  m12bandD.pop_front();
  m12bandD.push_back(percentage);
  ouTS.resize(m12bandD.size());
  std::copy(m12bandD.begin(),m12bandD.end(),ouTS.begin());
  dt12 = stride / static_cast<double>(ouTS.size()-1);
  mNoiseFloorBandD12.setData(t12, dt12, &ouTS[0], ouTS.size());
}

//======================================
//======================================  Write html Monitor status document.
void
NoiseFloorMonitor::DumpHTMLStatusPage(void)
{
  // Create a link to main page
  html::link linkToMainPage("Link to Monitor's Main Page","index.html");

  // Status Page
  // Set the current time
  char cbuf[1024];
  LocalStr(Now(), cbuf, "%M %d, %Y %02H:%02N");

  // Create html document
  html::document doc("Noise Floor Monitor Status Page");
  doc.setBackgroundColor(html::color("wheat"));

  // Write title
  html::text myTitle(std::string("Status page of the Noise Floor Tracker Monitor"));
  myTitle.setSize(html::size(2));
  html::block cb01("center");
  cb01.addObject(myTitle);
  doc.addObject(cb01);

  html::hline horizontalline;
  doc.addObject(horizontalline);

  html::text ln01(string("Last Noise Floor statistic at: ") + cbuf);
  //ln01.setSize(html::size(2));
  html::block lb01("left");
  lb01.lineBreak();
  lb01.addObject(ln01);
  doc.addObject(lb01);

  if  (mLockConditionFlag)
  {
    html::text ln02(std::string("Detector Lock Status: locked"));
    html::text ln03(std::string("Detector Lock Condition: ") + mLockConditionName);
    html::block lb02("left");
    lb02.lineBreak();
    lb02.addObject(ln02);
    lb02.lineBreak();
    lb02.addObject(ln03);
    doc.addObject(lb02);
  }
  else
  {
    html::text ln02(std::string("Detector Lock Status: not locked"));
    html::text ln03(std::string("Detector Lock Condition: ") + mLockConditionName);
    html::block lb02("left");
    lb02.lineBreak();
    lb02.addObject(ln02);
    lb02.lineBreak();
    lb02.addObject(ln03);
    doc.addObject(lb02);
  }
  
  // write threshold values
  std::ostringstream thrA;
  thrA << "Threshold band A: " << mStdDevA;
  std::ostringstream thrB;
  thrB << "Threshold band B: " << mStdDevB;
  std::ostringstream thrC;
  thrC << "Threshold band C: " << mStdDevC;
  std::ostringstream thrD;
  thrD << "Threshold band D: " << mStdDevD;
  html::text ln04(std::string(thrA.str()));
  html::text ln05(std::string(thrB.str()));
  html::text ln06(std::string(thrC.str()));
  html::text ln07(std::string(thrD.str()));
  html::block thrBlock("left");
  thrBlock.addObject(ln04);
  thrBlock.lineBreak();
  thrBlock.addObject(ln05);
  thrBlock.lineBreak();
  thrBlock.addObject(ln06);
  thrBlock.lineBreak();
  thrBlock.addObject(ln07);
  doc.addObject(horizontalline);
  doc.addObject(thrBlock);

  // Link back to main page
  html::block spb("left");
  spb.lineBreak();
  doc.addObject(spb);
  doc.addObject(horizontalline);
  doc.addObject(linkToMainPage);

  spb.lineBreak();

  // Write into file
  ofstream f(mHTMLStatusFile.c_str());
  html::writer w(f);
  doc.write(w);
}


//======================================
//======================================  Write html Monitor status document.
void
NoiseFloorMonitor::DumpHTMLErrorPage(void)
{
  // Auxiliary data
  char cbuf[1024];
  LocalStr(Now(), cbuf, "%M %d, %Y %02H:%02N");
  // Error Page
  // Create html document
  html::document doc("Noise Floor Monitor Error Page");
  doc.setBackgroundColor(html::color("wheat"));

  // Write title
  html::text myTitleErrPage(std::string("Error page of the Noise Floor Tracker Monitor"));
  myTitleErrPage.setSize(html::size(2));
  html::block cb01ErrPage("center");
  cb01ErrPage.addObject(myTitleErrPage);
  doc.addObject(cb01ErrPage);

  // Add an horizontal line
  html::hline horizontalline;
  doc.addObject(horizontalline);

  // Write out the lock sate
  if  (mLockConditionFlag)
  {
    html::text ln01ErrPage(std::string("Detector Lock Satus: locked"));
    html::text ln02ErrPage(std::string("Detector Lock Condition: ") + mLockConditionName);
    html::block lb01ErrPage("left");
    lb01ErrPage.lineBreak();
    lb01ErrPage.addObject(ln01ErrPage);
    lb01ErrPage.lineBreak();
    lb01ErrPage.addObject(ln02ErrPage);
    doc.addObject(lb01ErrPage);
    lb01ErrPage.lineBreak();
    lb01ErrPage.lineBreak();
  }
  else
  {
    html::text ln01ErrPage(std::string("Detector Lock Status: not locked"));
    html::text ln02ErrPage(std::string("Detector Lock Condition: ") + mLockConditionName);
    html::block lb01ErrPage("left");
    lb01ErrPage.lineBreak();
    lb01ErrPage.addObject(ln01ErrPage);
    lb01ErrPage.lineBreak();
    lb01ErrPage.addObject(ln02ErrPage);
    doc.addObject(lb01ErrPage);
    lb01ErrPage.lineBreak();
    lb01ErrPage.lineBreak();
  }

  // Process Error messages
  doc.addObject(horizontalline);
  gwd::Singleton* single = gwd::Singleton::exemplar();
  if (single->empty())
  {
    html::text ln03ErrPage(std::string("No error messages to be reported at: ") + cbuf);
    html::block lb03ErrPage("left");
    lb03ErrPage.lineBreak();
    lb03ErrPage.addObject(ln03ErrPage);
    doc.addObject(lb03ErrPage);
  }
  else
  {
    html::block eBlock("left");
    gwd::Singleton::Iterator p;
    for (p = single->begin(); p != single->end(); ++p)
    {
      html::text line(*(*p));
      eBlock.addObject(line);
      eBlock.lineBreak();
    }
    doc.addObject(eBlock);
  }
  
  // Link back to main page
  html::block spb("left");
  doc.addObject(spb);
  doc.addObject(horizontalline);
  html::link linkToMainPage("Link to Monitor's Main Page","index.html");
  doc.addObject(linkToMainPage);

  // Write into file
  std::ofstream ferr(mHTMLErrorFile.c_str());
  html::writer werr(ferr);
  doc.write(werr);

  // Clear message buffer
  single->ClearMessages();
}


//##########################################################################################################
//
//===================================== Auxiliary functions
//
//##########################################################################################################

//======================================
//======================================  Read frequencies and band with for the Notch filter
void ReadNotchFilterFreq(const std::string file,std::vector<double>& freq,std::vector<double>& freqbw)
{
  std::ifstream ifd(file.c_str());
  double val1,val2;
  double dummy;
  std::list<double> freqlt;
  std::list<double> freqbwlt;
  if (ifd.fail())
  {
    std::string errText = "can't open file" + file;
    gwd::Error(errText);
  }
  while (!ifd.eof())
  {
    ifd >> val1 >> val2 >> dummy;
    freqlt.push_back(val1);
    freqbwlt.push_back(val2);
  }
  ifd.close();

  // Copy into arrays
  freq.resize(freqlt.size());
  freqbw.resize(freqbwlt.size());
  size_t nn = 0;
  std::list<double>::iterator p,q;
  for (p = freqlt.begin(),q = freqbwlt.begin(); p != freqlt.end() && q != freqbwlt.end(); ++p,++q)
  {
    freq[nn] = *p;
    freqbw[nn] = *q;
    nn++;
  }
  freqlt.clear();
  freqbwlt.clear();
  ifd.close();
}


// Estimate mean value and standard deviation of a time series
inline void Estimator(const std::vector<double>& ts,double& mean,double& stdDev)
{
  const int lcSize = (int) ts.size();
  mean = double(0);
  for (int ii = 0; ii < lcSize; ii++)
  {
    mean += ts[ii];
  }

  mean /= static_cast<double>(lcSize);

  stdDev = double(0);
  for (int ii = 0; ii < lcSize; ii++)
  {
    stdDev += gwd::square(ts[ii]-mean);
  }

  stdDev /= static_cast<double>(lcSize-1);
  stdDev = std::sqrt(stdDev);

}


inline void MixWithOscilator(const double carrierFreq,const double sampleRate,std::vector<double>& signal)
{
  double tt = double(0);
  double dt = 1./sampleRate;
  double factor = 2.*gwd::PI*carrierFreq;
  for (int ii = 0; ii < (int)signal.size(); ii++)
  {
    signal[ii] *= std::cos(factor*tt);
    tt += dt;
  }
}


inline void GetTrend(const double sigma,std::vector<double>& ts,double& maxVal,double& percentage)
{
  const unsigned int size = static_cast<unsigned int>(ts.size());
  const double threshold = 3.*sigma;
  int counter = 0;
  maxVal = ts[0];

  for (unsigned int ii = 0; ii < size; ii++)
  {
    // count values above and below threshold
    if (ts[ii] > threshold || ts[ii] < -threshold) counter++;

    // get min. and max values
    maxVal = (maxVal > fabs(ts[ii])) ? maxVal : fabs(ts[ii]);
  }

  percentage = static_cast<double>(counter*100)/static_cast<double>(size);
}
