#include "SegGener.hh"
#include "ParseLine.hh"
#include "Segment.hh"
#include "html/color.hh"
#include "html/document.hh"
#include "html/table.hh"
#include "html/text.hh"
#include "html/writer.hh"
#include <unistd.h>
#include <sys/stat.h>
#include <fstream>

using namespace std;
using namespace html;

//======================================  Identify process
#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/SegGener/SegGener.cc,v 1.12 2006/07/18 07:52:13 jzweizig Exp $"
#define PIDTITLE  "Segment generation from an OSC condition"
#include "ProcIdent.hh"

//======================================  Generate a main routine.
EXECDAT(SegGener)

//======================================  Construct an On/Off segment
SegCond::SegCond(void)
  : mVersion(0), mLimit(0.0), mType(kUnknown), mPrevious(false), 
    mInSegment(false), mStart(0), mStop(0), mLiveTime(0.0), mNSegment(0)
{}

//======================================  Construct an On/Off segment
SegCond::SegCond(const std::string& group, int v, const std::string& on)
  : mGroup(group), mVersion(v), mLimit(0.0), mType(kOnCond), mOnCond(on),
    mPrevious(false), mInSegment(false), mStart(0), mStop(0), mCurrent(0), 
    mLiveTime(0.0), mNSegment(0)
{}

//======================================  Construct a stop/start segment
SegCond::SegCond(const std::string& group, int v, const std::string& on, 
	  const std::string& off)
  : mGroup(group), mVersion(v), mLimit(0.0), mType(kStartCond), mOnCond(on), 
    mOffCond(off), mPrevious(false), mInSegment(false), mStart(0), mStop(0), 
    mCurrent(0), mLiveTime(0.0), mNSegment(0)
{}

//======================================  Destroy the segment definition.
SegCond::~SegCond(void) {
}

//======================================  Evaluate the segment state
bool
SegCond::evaluate(OperStateCondList& l, const Time& t, Interval dT) {
    mPrevious = mInSegment;

    switch (mType) {
    case kOnCond:
      mInSegment = l.satisfied(mOnCond);
      break;

    case kStartCond:
      if (mPrevious && l.satisfied(mOffCond)) mInSegment = false;
      if (!mPrevious && l.satisfied(mOnCond)) mInSegment = true;
      break;

    default:
      mInSegment = false;
      break;
    }

    bool segComplete = mPrevious != mInSegment;
    if (!mCurrent) {
        mStart = t;
	mStop  = t;
	segComplete = false;
    } else if (segComplete) {
        if (mPrevious) mStop  = t;
	else           mStart = t;
	if (mStart == mStop) segComplete = false;
    }

    //------------------------------------  Calculate time in segment
    mCurrent      = t + dT;
    Interval delt = mInSegment ? mCurrent-mStart : mCurrent-mStop;
    segComplete  |= (mLimit != Interval(0.0) && delt >= mLimit);
    if (mInSegment) mLiveTime += dT;
    return segComplete;
}

//======================================  Make a Segment
//
//   The segment actions should go as follows:
//
//   Prev  This  begin   end    action         comment
//    on    on   start  current start=current  Segment limit
//    on    off  start  stop    start=current  transition->off
//    off   on   stop   start   stop =current  transition->on
//    off   off  stop   current stop =current  Segment limit 
//
trig::Segment
SegCond::makeSegment(void) {
    Time tBegin, tEnd;
    if (mPrevious) {
        tBegin = mStart;
	tEnd   = mStop;
	mStart = mCurrent;
    } else {
        tBegin = mStop;
	tEnd   = mStart;
	mStop  = mCurrent;
    }
    if (mPrevious == mInSegment) tEnd = mCurrent; 

    trig::Segment s(mGroup, mVersion, tBegin, tEnd);
    if (!mPrevious) s.setActivity(0);
    s.setIfos(mIfo.c_str());
    s.setComment(mComment);
    mNSegment++;
    return s;
}

//======================================  Reset segment state.
void 
SegCond::reset(void) {
    mCurrent = Time(0);
}

//======================================  Set the segment list file.
void 
SegCond::setComment(const std::string& comment) {
    mComment = comment;
}

//======================================  Set the segment list file.
void 
SegCond::setIfo(const std::string& ifos) {
    mIfo = ifos;
}

//======================================  Set the maximum segment time.
void 
SegCond::setLimit(Interval dT) {
    mLimit = dT;
}

//======================================  Set the segment list file.
void 
SegCond::setSegFile(const std::string& file) {
    const char* htmdir = getenv("DMTHTMLOUT");
    if (file.empty() || file[0] == '.' || file[0] == '/' || !htmdir) {
        mSegFile = file;
    } else {
        mSegFile  = htmdir;
	mSegFile += "/";
	mSegFile += file;
    }

    //----------------------------------  Complain if the file can't be written
    if (!mSegFile.empty()) {
        if (access(mSegFile.c_str(), W_OK) && !access(mSegFile.c_str(), F_OK)){
	  cout << "Can't write file: " << mSegFile 
	       << " disabling text file output for segment: " << mGroup 
	       << endl;
	  mSegFile.clear();
	}
    }
}


//======================================  File and Date
static void
FileAnDate(std::ostream& out, const std::string& file) {
    struct stat statbuf;
    out << file << " [" << endl;
    stat(file.c_str(), &statbuf);
    Time tMod = fromUTC(statbuf.st_mtime);
    char datime[32];
    out << TimeStr(tMod, datime, "%3M %d, %Y %02H:02M %Z") << "]";
}

//======================================  Monitor constructor.
SegGener::SegGener(int argc, const char* argv[]) 
  : DatEnv(argc, argv), mConfig("SegGener.conf"), mOSC(mIn), mRecord(false), 
    mFirst(0), mLast(0)
{
    //----------------------------------  Parse the argument list
    bool syntax = false;

    for (int i=1 ; i<argc ; i++) {
        string argi(argv[i]);
	if (argi == "-conf") {
	    mConfig = argv[++i];
	} else if (argi == "-htfile") {
	    mHtFile = argv[++i];
	} else if (argi == "-osc") {
	    mOSCfile = argv[++i];
	} else if (argi == "+seg") {
	    mRecord = true;
	} else if (isDatEnvArg(argv[i])) {
	    i++;
	} else {
	    cerr << "Unrecognized argument: " << argi << endl;
	    syntax = true;
	    break;
	}
    }

    //----------------------------------  Check syntax
    if (mOSCfile.empty()) {
        cerr << "OSC file not specified!" << endl;
	syntax = true;
    }

    if (syntax) {
        cerr << "Syntax error discovered in command. Syntax is:" << endl;
        cerr << "SegGener [-conf <config-file>] -osc <OSC-file> [+seg]" 
	     << endl;
	finish();
	return;
    }

    //----------------------------------  Configure the OperStateCondList
    cout << "Reading OSC file: ";
    FileAnDate(cout, mOSCfile);
    cout << endl;
    mOSC.readConfig(mOSCfile);

    //----------------------------------  Read in the monitor configuration
    syntax = readConfig(mConfig.c_str());

    //----------------------------------  Invent an html file name
    if (mHtFile.empty()) {
        const char* d = getenv("DMTHTMLOUT");
	if (d) mHtFile = string(d) + "/";
	mHtFile += "Segment_List.html";
    }

    //----------------------------------  Bail out if there was an error.
    if (syntax) {
        cerr << "Syntax error found in file: " << mConfig 
	     << ". SegGener will terminate." << endl;
	finish();
	return;
    }
    getDacc().setIgnoreMissingChannel(true);
}

//======================================  Read the configuration file
bool
SegGener::readConfig(const char* cfile) {

    //----------------------------------  Open the configuration file.
    ParseLine pl(cfile);
    if (!pl.isOpen()) {
        cerr << "Unable to open configuration file: " << cfile << endl;
	return true;
    }

    //----------------------------------  Record the configuration
    cout << "Reading SegGener configuration file: ";
    FileAnDate(cout, cfile);
    cout << endl;

    //----------------------------------  Read in the configuration.
    bool syntax = false;
    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
        if (nArg == 0) continue;
	int iArg = 3;
	SegCond sc;
	if (nArg < 3) {
	    cerr << "Error: Insufficient arguments on line." << endl;
	    syntax = true;
	} else if (nArg == 3 || *(pl[3]) == '-') {
	    sc = SegCond(pl[0], pl.getInt(1), pl[2]);
	    if (!OSCexists(pl[2])) {
	        cerr << "On condition (" << pl[2] << ") for segment group "
		     << pl[0] << " doesn't exist or isn't valid." << endl;
		syntax = true;
	    } else if (Debug()) {
	        cout << "Defined segment: " << pl[0] << " version: "
		     << pl.getInt(1) << " condition: " << pl[2] << endl;
	    }
	} else {
	    if (!OSCexists(pl[2])) {
	        cerr << "Start condition (" << pl[2] << ") for segment group "
		     << pl[0] << " doesn't exist or isn't valid." << endl;
		syntax = true;
	    } else if (!OSCexists(pl[3])) {
	        cerr << "Stop condition (" << pl[3] << ") for segment group "
		     << pl[0] << " doesn't exist or isn't valid." << endl;
		syntax = true;
	    } else if (Debug()) {
	        cout << "Defined segment: " << pl[0] << " version: "
		     << pl.getInt(1) << " start condition: " << pl[2] 
		     << " stop condition: " << pl[3] << endl;
	    }
	    sc = SegCond(pl[0], pl.getInt(1), pl[2], pl[3]);
	    iArg = 4;
	}
	for (int i=iArg ; i<nArg ; i++) {
	    string argi(pl[i]);
	    if (argi == "-comment") {
	        sc.setComment(pl[++i]);
	    } else if (argi == "-file") {
	        sc.setSegFile(pl[++i]);
	    } else if (argi == "-ifo") {
	        sc.setIfo(pl[++i]);
	    } else if (argi == "-limit") {
	        sc.setLimit(pl.getDouble(++i));
	    } else {
	        cerr << "Unrecognized argument in configuration file: " 
		     << argi << endl;
		syntax = true;
	    }
	}
	if (sc.refGroup()[2] == ':') {
	    if (sc.refIfo().empty()) sc.setIfo(sc.refGroup().substr(0,2));
	    sc.refGroup().erase(0,3);
	} else if (sc.refStartCond()[2] == ':' && sc.refIfo().empty()) {
	    sc.setIfo(sc.refStartCond().substr(0,2));
	}
	mList.push_back(sc);
    }
    return syntax;
}

//======================================  SegGener monitor termination.
SegGener::~SegGener(void) {

    //----------------------------------  Write out any pending segments.
    for (list_iter i=mList.begin() ; i != mList.end() ; i++) {
	if (i->testInSegment()) {
	    trig::Segment s(i->getGroup(), i->getVersion(), 
			    i->getStartTime(), i->getEndTime());
	    if (mRecord) {
	        sendSegment(s);
	    } else {
	        cout << "Segment group: " << s.getGroup() << " version: "
		     << s.getVersion() << " Started: " << s.getStartTime()
		     << " and ended: " << s.getEndTime() << endl;
	    }
	    i->reset();
	}
    }

    //----------------------------------  Clear the segment definitions.
    mList.clear();
}

//======================================  Process a stride, generate segments
void 
SegGener::ProcessData(void) {
    Time     t  = getDacc().getFillTime();
    Interval dT = getDacc().getStride();
    if (!mFirst) mFirst = t;
    for (list_iter i=mList.begin() ; i != mList.end() ; i++) {
        if (i->evaluate(mOSC, t, dT)) {
	    if (Debug()) {
	        if (i->testInSegment()) cout << "In " << i->getGroup() 
					     << " segment at: " << t << endl;
		else                    cout << "Out of " << i->getGroup() 
					     << " segment at: " << t << endl;
	    }
	    trig::Segment s = i->makeSegment();
	    if (mRecord) {
	        sendSegment(s);
	    } else {
	        cout << "Segment group: " << i->getGroup() << " version: "
		     << i->getVersion() << " Started: " << s.getStartTime()
		     << " and ended: " << s.getEndTime() << endl;
	    }
	    const char* segFile = i->getSegFile();
	    if (segFile && *segFile && s.getActivity() > 0) {
	        ofstream otxt(segFile, ios::app);
		if (otxt.is_open()) {
		    otxt << i->getNSegment() << " " << s.getStartTime().getS() 
			 << " " << s.getEndTime().getS() << " " 
			 << long(s.getEndTime() - s.getStartTime()) << endl;
		}
	    }
	}
    }
    mLast = t + dT;
    if (t.getS() % 60 == 13) writeTable();
}

//======================================  Build an html page
void
SegGener::writeTable(void) const {
    char line[128];

    //----------------------------------  Define the table.
    document doc("Segment table");
    doc.setBackgroundColor(color("white"));
    doc.addObject(text(TimeStr(Now(), line, 
	       "Report generated:        %Y-%02m-%02d %H:%N:%S (GPS %s)")));
    doc.lineBreak();
    doc.addObject(text(TimeStr(mFirst, line, 
	       "Data processing started: %Y-%02m-%02d %H:%N:%S (GPS %s)")));
    doc.lineBreak();

    std::ostringstream out;
    out.str("Segment Configuration file [date]: ");
    FileAnDate(out, mConfig);
    doc.addObject(text(out.str()));
    doc.lineBreak();

		  
    out.str("OSC condition file [date]: ");
    FileAnDate(out, mOSCfile);
    doc.addObject(text(out.str()));
    doc.lineBreak();

    table* tab = dynamic_cast<table*>(&doc.addObject(table("Segment Types")));
    tab->addColumn("Segment \nGroup");
    tab->addColumn("In \nSegment");
    tab->addColumn("Start Condition");
    tab->addColumn("Stop Condition");
    tab->addColumn("Live (%)");
    tab->addColumn("Segments\nGenerated");
    tab->addColumn("Time\nLimit");

    //----------------------------------  Fill the table.
    double denom = double(mLast - mFirst);
    if (denom <= 0) denom = 1.0;
    denom /= 100.0;  // scale to percent.
    for (const_iter i=mList.begin() ; i != mList.end() ; i++) {
        int j = tab->addRow();
	tab->insertData(j, 0, i->getGroup());
	if (i->testInSegment()) tab->insertData(j, 1, "Yes");
	else                    tab->insertData(j, 1, "No");
	tab->insertData(j, 2, i->getStartCond());
	tab->insertData(j, 3, i->getStopCond());
	tab->insertData(j, 4, double(i->getLiveTime())/denom);
	tab->insertData(j, 5, long(i->getNSegment()));
	tab->insertData(j, 6, double(i->getLimit()));
    }

    //----------------------------------  Write the html page.
    ofstream s(mHtFile.c_str(), ios::out);
    writer w(s);
    doc.write(w);
    s.close();
}
