/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: rockIFO							*/
/*                                                         		*/
/* Module Description: tool to play back LIGO data channels in          */
/* real time on Linux machines                                          */
/*									*/
/*                                                         		*/
/* Module Arguments: none				   		*/
/*                                                         		*/
/* Revision History:					   		*/
/* Rel   Date       Programmer  	Comments			*/
/* 0.1	 12May2005  S. Ballmer   	First release                   */
/*                                                         		*/
/* Documentation References:						*/
/*	Man Pages: none					                */
/*	References: none						*/
/*                                                         		*/
/* Author Information:							*/
/* Name          Telephone       Fax             e-mail 		*/
/* Stefan Ballmer (617)253-0203	  -              sballmer@ligo.mit.edu  */
/* Daniel Sigg    (509)372-8336  (509) 372-2178  sigg_d@ligo.mit.edu	*/
/*                                                         		*/
/* Code Compilation and Runtime Specifications:				*/
/*                                                         		*/
/* Code Standards Conformance:						*/
/*									*/
/* Known Bugs, Limitations, Caveats:					*/
/*								 	*/
/*									*/
/*                                                         		*/
/*                             LIGO					*/
/*                                                         		*/
/*        THE LASER INTERFEROMETER GRAVITATIONAL WAVE OBSERVATORY.	*/
/*                                                         		*/
/*                     (C) The LIGO Project, 1996.			*/
/*                                                         		*/
/*                                                         		*/
/* California Institute of Technology			   		*/
/* LIGO Project MS 51-33				   		*/
/* Pasadena CA 91125					   		*/
/*                                                         		*/
/* Massachusetts Institute of Technology		   		*/
/* LIGO Project MS 20B-145				   		*/
/* Cambridge MA 01239					   		*/
/*                                                         		*/
/* LIGO Hanford Observatory				   		*/
/* P.O. Box 1970 S9-02					   		*/
/* Richland WA 99352					   		*/
/*                                                         		*/
/* LIGO Livingston Observatory		   				*/
/* 19100 LIGO Lane Rd.					   		*/
/* Livingston, LA 70754					   		*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#include <TSeries.hh>
#include <FilterBase.hh>
#include <FilterDesign.hh>
#include <tconv.h>
#include <laudio.hh>
#include "rockaudio.hh"
#include "resample.hh"
#include "LiveDataServer.hh"
#include <iostream>

using namespace std;

/*----------------------------------------------------------------*\
  Main function
\*----------------------------------------------------------------*/
int 
main(int argc, const char *argv[]) {

//------------------------------ local variables ------------------
  LiveDataServer lds;
  const LiveData* chanData;
 
  int chanNum=1;					// number of channels
    const char*   ndsname = NULL;
    const char*   measChan = "H1:LSC-DARM_ERR";
    const char*   filterStr = "";
    bool          suspendFlag=false;
    double        settleTime=0; //seconds
    bool          debug = false;
    int		  errflag = 0;	/* error flag */
    double	  timeout = -1.0;	/* timeout */
    double        soundDelay=0;
    double        biggestSoundDelay=0;
    double        maxSoundDelay=0.25;
    double mVolume =0.5;
    int    mMode =1;
    bool   mMute =0;
    double mfSample=14200;
    bool autogain=false;
    double AGain=1.0;
   
   
//---------------------------------  Get the arguments ------------------
     for (int i=1 ; i<argc ; i++) {
        if (!strcmp(argv[i], "nds")) {
	    ndsname = argv[++i];
        } else if (!strcmp(argv[i], "meas") ||
	           !strcmp(argv[i], "channel") ||
	           !strcmp(argv[i], "listento")) {
	    measChan = argv[++i];
        } else if (!strcmp(argv[i], "filter")) {
	    filterStr = argv[++i];
        } else if (!strcmp(argv[i], "settle")) {
            settleTime  = atof(argv[++i]);
        } else if (!strcmp(argv[i], "autogain")) {
	    autogain=true;
        } else if (!strcmp(argv[i], "Volume") ||
	           !strcmp(argv[i], "volume")) {
	    mVolume=atof(argv[++i]);
        } else if (!strcmp(argv[i], "Mode")) {
	    mMode=atoi(argv[++i]);
        } else if (!strcmp(argv[i], "Mute")) {
	    mMute=(atoi(argv[++i]));
        } else if (!strcmp(argv[i], "fsample")) {
	    mfSample=(atof(argv[++i]));
        } else if (!strcmp(argv[i], "-t") ||
	           !strcmp(argv[i], "timeout")) {
                  timeout = atof (argv[++i]);
        } else if (!strcmp(argv[i], "help")   || 
	           !strcmp(argv[i], "-help")  ||
		   !strcmp(argv[i], "--help") ||
		   !strcmp(argv[i], "-h")     ||
		   !strcmp(argv[i], "-?")     ||
		   !strcmp(argv[i], "?")) {	  
            /* help */
                  errflag = 1;
	} else if (!strcmp(argv[i], "debug")) {
	    debug = true;
	} else {
  	    measChan = argv[i];
	}
    }

      /* help */
      if (errflag || argc==1) {
         printf ("Usage: %s [options] \n"
	        "       nds      'nds-ip-address': nds server\n"
	        "       channel  'channel': channel to listen to (channel can be omitted)\n"
	        "       filter   'filter design string (same syntax as in Foton)' \n"
		"       autogain  use automatic gain adjustment feature\n"
		"       volume   'volume' set the audio volume\n"
		"                 if autogain is specified: volume = DAC range / peak\n"
		"                 otherwise volume is just a gain\n"
		"       fsample  'sampling rate of the sound card. Default is 14200. Change if sound is not continuos.\n"
		"       timeout  'secs' set time after which to quit in seconds\n"
		"       settle   'time': settling time in seconds - default is 0sec\n"
		"       debug:    enable debug flag for DAQSocket class\n"
                "       help      help\n\n"
		"Examples:\n"
		"       1) Play H0:PEM-LVEA_MIC:\n"
		"          rockIFO nds red.ligo-wa.caltech.edu H0:PEM-LVEA_MIC autogain\n"
                "       2) Play H1:LSC-DARM_ERR:\n"
		"          rockIFO nds red.ligo-wa.caltech.edu H1:LSC-DARM_ERR volume 0.1 \\\n"
		"                  autogain filter \"butter('HighPass',8,100)notch(343,10,20)\"\n",argv[0]);
         return 1;
      }

      if (timeout == 0) {
         return 0;
      }
      
      if(!autogain) {
        AGain=mVolume;
      }

//---------------------------------- wait for the system to settle ------
   tainsec_t SettleT0 = TAInow();
   while ((double)((TAInow()-SettleT0)/1E9)<=settleTime);

//--------------------------------- Open a daqd socket ------------------
    if(debug) cout << "Opening a socket to ND server on node " << ndsname << endl;
 // initialize live data server, exit on error
  printf("Connecting to NDS server...\n");
  if( !lds.initNDS(ndsname) )
    exit(1);					// ======== EXIT

  chanData = lds.initChannels(&measChan, chanNum);
  if( chanData == NULL || !lds.initDataFlow() )
    exit(1);					// ======== EXIT
    
//---------------------------------- Add channels to DAQSocket -----------
   
      // start time
      tainsec_t t0 = TAInow();
//---------------------------------- Set up sound stuff -----------
      double t = 0;
      double prev = 0;
      double dt;
      int soundCycle=-1;
      bool soundFirst=true;
   
   
//---------------------------------- Set up audio -----------
     LAudio mPort;
     FilterDesign* filterdesign;
     FilterDesign* filterdesign2;
     Pipe* filter;
     Pipe* filter2;
     double fsample=mfSample;
     filterdesign = new FilterDesign(chanData[0].daq.mRate);
     filterdesign2 = new FilterDesign(fsample);
     //filterdesign->wizard();
     if( !filterdesign->filter(filterStr) ) cout <<filterStr<<" is no valid filter"<<endl;
     double fCut= chanData[0].daq.mRate < fsample ? 0.45*chanData[0].daq.mRate : 0.45*fsample; 
     filterdesign->butter(kLowPass,4,fCut);
     //filterdesign2->butter(kLowPass,2,fCut);
     filter=filterdesign->release();
     filter2=filterdesign2->release();
     int dev=-1;
     TSeries ts;
     TSeries tsf;
     TSeries tsds;
     TSeries tsdsf;
     float buf2[16384];
     float bufT[16384];
     int Tlen,Tlen2;
     ResampleT* rsd=SetResample(int(chanData[0].daq.mRate),int(fsample));
     //int toggle=0;
     //int LTlen2=888;

//---------------------------------- request the data -----------
   //int erno = nds.RequestOnlineData(true,-1);
   //if(erno) cout << "Server error #"<<erno << endl;
   
//---------------------------------- local variables for demodulation -----------
   double *fbuf;   
//----------------------------- start play loop --------------------
      printf("Playing %s\n",measChan);
      do {
         // wait a little while
         //nanosleep (&waittime, 0);
         // get error signal
    if( !lds.getData() )
      exit(1);					// ======== EXIT

 

     {
       fbuf=(chanData[0].data);
       
       int Nsamp= chanData[0].daq.mRate/16;
       int mOff=0;
//----------------------------- loop through received data --------------------
       //cout << "converting to TSeries"<<endl;
       ts= TSeries(0,1.0/chanData[0].daq.mRate,Nsamp,fbuf+mOff);
       //cout << "applying filter"<<endl;
       tsf=filter->apply(ts);
       Tlen=tsf.getData(16384,bufT);
       
       Resample(Tlen,bufT,&Tlen2,buf2,rsd,AGain);
       
       tsds=TSeries(0,1.0/fsample,Tlen2,buf2);
       tsdsf=filter2->apply(tsds);
       double mx=tsdsf.getMaximum();
       double mn=tsdsf.getMinimum();
       double amx= abs(mn) > abs(mx) ? abs(mn) : abs(mx);
       //cout << "TSeries ready, "<<1.0/ts.getTStep() <<" / "<<1.0/tsdsf.getTStep()<<endl;
       if(debug) cout << "TSeries ready, "<<ts.getNSample() <<" / "<<tsdsf.getNSample()<<endl;
//---------------------------- recalculate the gain --------------------
      if(autogain && amx>0.0) {
        double TAG=mVolume/amx;
	if(TAG<0.01) {
	  AGain=TAG;
	} else {
          AGain*=0.99+ TAG*0.01;
	}
	if(AGain>1e16) AGain=1;
	if(AGain<1e-16) AGain=1;
	if(debug) {
	  cout << mVolume/amx <<endl;
	  cout << "New Gain is "<<AGain<<endl;
	}
      }
//----------------------------- start sounding part ----------------     
         // get time
	 if(soundFirst) {
	   soundFirst=false;
	   t0 = TAInow();
	   prev=0;
	 }
         t = (TAInow() - t0) / 1E9;
	 soundCycle++;
         dt = t - prev;
     //check whether we keep up with real time
     soundDelay=t - (double)soundCycle/16.0;
     if(biggestSoundDelay < soundDelay) biggestSoundDelay=soundDelay;
     if(soundDelay<maxSoundDelay && !suspendFlag) {
       prev = t;
       // play the sound
       if(debug) cout << "sending audio now; Delay = " << soundDelay <<endl;
       playTSeries(&dev,tsdsf,0);
     } else {
       if(!suspendFlag)
         printf("Can't keep up with realtime data stream, skipping audio output\n");
       }
      }
         // loop until timeout
      } while (((timeout < 0) || (t < timeout)) );
//----------------------------- clean up --------------------
      printf("done\n");
   exit(0);
}

