//
//    Sandbox data access implementation
//
#if defined(sun) && !defined(__EXTENSIONS__)
#define __EXTENSIONS__
#endif
#include <strings.h>

#include "DaccIn.hh"
#include "FSeries.hh"
#include "TSeries.hh"
#include "framecpp/FrameH.hh"
#include "framecpp/FrHistory.hh"
#include "general/gpstime.hh"
#include "framecpp/IFrameStream.hh"
#include "FrVectRef.hh"
#include "FrStatDataRef.hh"
#include "DVecType.hh"
#ifndef DMTOFFLINE
#include "iSMbuf.hh"
#endif
#include <iostream>
#include <fstream>

using FrameCPP::FrAdcData;
using FrameCPP::FrameH;
using FrameCPP::IFrameStream;
using FrameCPP::FrProcData;
using FrameCPP::FrRawData;
using FrameCPP::FrSimData;
using FrameCPP::FrStatData;

using namespace std;

//======================================  Time object conversion functions.
inline Time
getTime(const General::GPSTime& t) {
    return Time(t.getSec(), t.getNSec());
}

inline General::GPSTime
getGPS(const Time t) {
    return General::GPSTime(t.getS(), t.getN());
}

//======================================  Default (Null) constructor.
DaccIn::DaccIn(void) 
  : mDebug(0), mT0(0), mFramesInFile(0), mFrameNo(0), mTOCMode(s_TOC),
    mSbuf(0), mStream(0), mReader(0), mFrame(0), mTotalFrames(0)
{
    mSource = s_none;
}

//======================================  Construct and connect to source.
DaccIn::DaccIn(IFrameStream* reader) 
  : mDebug(0), mT0(0), mFramesInFile(0), mFrameNo(0), mSbuf(0), mStream(0),
    mReader(reader), mFrame(0), mTotalFrames(0)
{
    mSource = s_none;
    setTOCMode(false);
}

//======================================  Destructor.
DaccIn::~DaccIn(){
    close();
}

//======================================  Specify number of online buffers
void
DaccIn::setBuffer(int N) {
#ifndef DMTOFFLINE
    if (!isOnline()) return;
    dynamic_cast<iSMbuf*>(mSbuf)->setBCount(N);
#endif
}

//======================================  Set the TOC mode
void
DaccIn::setTOCMode(bool toc) {
    if (toc) mTOCMode = s_TOC;
    else     mTOCMode = s_noTOC;
}

//======================================  Get Time span of current frame.
Interval
DaccIn::getDt(void) const {
    if (!haveFrame()) return Interval(0.0);
    return Interval(mFrame->GetDt());
}

//======================================  Build the Frame name.
string
DaccIn::getFrameID(void) {

    //----------------------------------  Make sure we have a frame.
    if (!haveFrame()) return string("");

    //----------------------------------  Get the frame time, define the Epoch
    if (mIDFormat.empty()) {
        const char* site = getenv("LIGOSITE");
	if (!site) {
	    mIDFormat = "X-%s.F";
	} else if (!strcmp(site, "LHO")) {
	    mIDFormat = "H-%s.F";
	} else if (!strcmp(site, "LLO")) {
	    mIDFormat = "L-%s.F";
	} else if (!strcmp(site, "CIT")) {
	    mIDFormat = "C1-%s.F";
	} else {
	    mIDFormat = "X-%s.F";
	}
    }

    //----------------------------------  Construct a valid name.
    char tstr[40];
    TimeStr(::getTime(mFrame->GetGTime()), tstr, mIDFormat.c_str());
    return string(tstr);
}

//======================================  Read in static data
int 
DaccIn::getStaticData(const std::string& name,  const std::string& det, 
		      const Time& startGps, const Time& endGps) {
    if (!mReader) return 0;
    mReader->ReadFrStatData(name, det, getGPS(startGps), getGPS(endGps), 
			    FrStatData::LATEST_VERSION, mStatData);
    int N = mStatData.size();
    for (int i=0; i<N; i++) mStatData[i];
    if (mDebug) {
        cerr << "Number of FrStatdata structures: " << N << endl;
	if (mDebug > 1) {
	    cerr << "Name  gps-start gps-stop version" << endl;
	    for (int i=0; i<N; i++) {
	        const FrameCPP::FrStatData* p=mStatData[i];
		cerr << p->GetName()    << " " << p->GetTimeStart() << " "
		     << p->GetTimeEnd() << " " << p->GetVersion()   << endl;
	    }
	}
    }
    return N;
}

//======================================  Get a static FSeries
FSeries
DaccIn::getStaticFSeries(const std::string& name, const Time& gps) {
    FrStatDataRef p = findStat(name, gps);
    if (!p.null()) {
	FrVectRef v = p.getFrVect();
	double   f0 = v.getDimX0(0);
	double   dF = v.getDimDx(0);
	Interval dT = double(p.getEndTime() - p.getStartTime());
	return FSeries(f0, dF, p.getStartTime(), dT, v.getDVector());
    }
    return FSeries();
}

//======================================  Get a static TSeries
TSeries
DaccIn::getStaticTSeries(const std::string& name, const Time& gps) {
    FrStatDataRef p = findStat(name, gps);
    if (!p.null()) {
	FrVectRef  v = p.getFrVect();
	Time      t0 = p.getStartTime() + Interval(v.getDimX0(0));
	Interval  dT = v.getDimDx(0);
	return TSeries(t0, dT, v.getDVector());
    }
    return TSeries();
}

//======================================  Connect to a data source.
bool 
DaccIn::getHistory(const std::string& name, Time& t, std::string& com) const {
    if (haveFrame()) {
        for (FrameH::history_iterator itr=mFrame->RefHistory().begin();
	     itr != mFrame->RefHistory().end();
	     itr++) {
	    if ( (*itr)->GetName() == name) {
	        t   = (*itr)->GetTime();
		com =  (*itr)->GetComment();
		return true;
	    }
	}
    }
    return false;
}

//======================================  Connect to a data source.
int
DaccIn::open(void) {
    if (isOpen()) return 0;
    if (!mStream) {
        const char* Source = getFile();
	if (!Source) {
	    cout << "No more requested files" << endl;
	    return -1;
	}
	int rc = openFile(Source); // prints error and pops file name
	if (rc) return rc;
    }
    int rc = openReader();
    if (rc) {
        cerr << "Unable to open reader on file: " << getFile() << endl;
        closeFile();               // discard file.
    }
    return rc;
}

//======================================  Disconnect from the data source.
void
DaccIn::close(void) {
    endFrame();
    closeReader();
    closeFile();
}

//======================================  Set the default stride.
void 
DaccIn::setDebug(int level) {
    mDebug = level;
}

//======================================  Clean up when frame no longer needed
void 
DaccIn::endFrame(void) {
    if (!haveFrame()) return;

    //-------------------------- Delete the frame and bump the frame number.
    deleteFrame();

#ifndef DMTOFFLINE
    //-------------------------- Note that the following code attempts to free
    //                           a partiton buffer for online code. It isn't
    //                           really correct in that the reader is deleted
    //                           and buffer freed for non-TOC access after the
    //                           first buffer.
    if (isOnline()) {
        if (mTOCMode != s_TOC || mFramesInFile <= mFrameNo) {
	    dynamic_cast<iSMbuf*>(mSbuf)->relse();
	    closeReader();
	}
    }
#endif
}

//======================================  Get the next frame  
int
DaccIn::nextFrame(void) {
    if (haveFrame()) deleteFrame();

    while (!haveFrame()) {
        if (!isOpen() && open()) return -4;
	int rc = readFrame();
	if (!rc) {
	    continue;
	} else if (!isOnline()) {
	    close();
	} else if (mTOCMode == s_TOC) {
#ifndef DMTOFFLINE
	    dynamic_cast<iSMbuf*>(mSbuf)->relse();
#endif
	    closeReader();
	}
    }

    mT0 = ::getTime(mFrame->GetGTime());
    const FrRawData* raw = getFrame()->GetRawData();
    if (raw) mAdcIter = raw->RefFirstAdc().begin();
    mProcIter = getFrame()->RefProcData().begin();
    mSimIter  = getFrame()->RefSimData().begin();
    mTotalFrames++;
    return 0;
}

//======================================  Wait for an online frame to arrive
bool
DaccIn::waitData(bool nowait) const {
#ifndef DMTOFFLINE
    if (!isOnline()) return true;
    return dynamic_cast<iSMbuf*>(mSbuf)->waitBuf(nowait);
#else
    return true;
#endif
}

//======================================  Search entire list for an adc
const FrAdcData* 
DaccIn::findAdc(const std::string& name) const {
    const FrRawData* raw = getFrame()->GetRawData();
    if (raw) mAdcIter = raw->RefFirstAdc().begin();
    return nextAdc(name);
}

//======================================  Search list from previous point
const FrAdcData* 
DaccIn::nextAdc(const std::string& name) const {
    const FrRawData* raw = mFrame->GetRawData();
    if (mDebug > 3) cout << "Searching for Adc: " << name << "...";
    switch (mTOCMode) {
    case s_TOC:
        if (!raw) return readAdc(name);
        mAdcIter = raw->RefFirstAdc().begin();
	mAdcIter = raw->RefFirstAdc().find(name, mAdcIter);
	if (mAdcIter == raw->RefFirstAdc().end()) return readAdc(name);
	break;
    case s_noTOC:
	if (!raw) {
	    if (mDebug) cerr << "No Raw data in frame!" << endl;
	    return 0;
	}
	mAdcIter = raw->RefFirstAdc().find(name, mAdcIter);
	if (mAdcIter == raw->RefFirstAdc().end()) {
	    if (mDebug > 3) cout << " Not found!" << endl;
	    return 0;
	}
    }
    if (mDebug > 3) cout << " Found!" << endl;
    return *mAdcIter;
}

//======================================  Search entire list for ProcData
const FrProcData* 
DaccIn::findProc(const std::string& name) const {
    if (!mFrame) return 0;
    mProcIter = mFrame->RefProcData().begin();
    return nextProc(name);
}

//======================================  Search list from previous point
const FrProcData* 
DaccIn::nextProc(const std::string& name) const {
    if (mDebug > 3) cout << "Searching for FrProc: " << name << " ...";
    if (!mFrame) {
        if (mDebug) cout << "No frame available!" << endl;
        return 0;
    }
#ifdef ENABLE_DEPRICATED
    const FrameCPP::FrameH::ProcDataContainer& pCntr = mFrame->RefProcData();
#else
    const FrameCPP::FrameH::procData_type& pCntr = mFrame->RefProcData();
#endif
    switch (mTOCMode) {
    case s_TOC:
        mProcIter = pCntr.begin();
	mProcIter = pCntr.find(name, mProcIter);
	if (mProcIter == pCntr.end()) return readProc(name);
	break;
    case s_noTOC:
	mProcIter = pCntr.find(name, mProcIter);
	if (mProcIter == pCntr.end()) {
	    if ( mDebug > 3) cout << " Not found!" << endl;
	    return 0;
	}
    }
    if ( mDebug > 3) cout << " Found!" << endl;
    return *mProcIter;
}

//======================================  Search entire list for ProcData
const FrSimData* 
DaccIn::findSim(const std::string& name) const {
    if (!mFrame) return 0;
    mSimIter = mFrame->RefSimData().begin();
    return nextSim(name);
}

//======================================  Search list from previous point
const FrSimData* 
DaccIn::nextSim(const std::string& name) const {
    if (mDebug > 3) cout << "Searching for FrSim: " << name << " ...";
    if (!mFrame) {
        if (mDebug) cout << "No frame available!" << endl;
        return 0;
    }
#ifdef ENABLE_DEPRICATED
    const FrameCPP::FrameH::SimDataContainer& pCntr = mFrame->RefSimData();
#else
    const FrameCPP::FrameH::simData_type&     pCntr = mFrame->RefSimData();
#endif
    switch (mTOCMode) {
    case s_TOC:
        mSimIter = pCntr.begin();
	mSimIter = pCntr.find(name, mSimIter);
	if (mSimIter == pCntr.end()) return readSim(name);
	break;
    case s_noTOC:
	mSimIter = pCntr.find(name, mSimIter);
	if (mSimIter == pCntr.end()) {
	    if ( mDebug > 3) cout << " Not found!" << endl;
	    return 0;
	}
    }
    if ( mDebug > 3) cout << " Found!" << endl;
    return *mSimIter;
}

//======================================  Find a specified static structure
FrStatDataRef 
DaccIn::findStat(const std::string& name, const Time& gps, int vsn) {
    if (mDebug > 1) cerr << "Look for FrStatData: " << name << " version: "
			 << vsn << " for GPS: " << gps.getS() << "... ";
    FrStatDataRef psav;
    Time tLast(0);
    int  pvsn(-1);
    int N = mStatData.size();
    for (int i=0; i<N; i++) {
        FrStatDataRef p(mStatData[i]);
	if (name != p.getName()) continue;
	int cvsn = p.getVersion();
	if ((vsn < 0 && cvsn >= pvsn) || cvsn == vsn) {
	    if (!gps && p.getStartTime() > tLast) {
	        psav  = p;
		pvsn  = cvsn;
		tLast = p.getStartTime();
	    } else if (gps >= p.getStartTime() && gps < p.getEndTime()) {
	        psav = p;
		pvsn = cvsn;
	    }
	}
    }
    if (mDebug > 1) {
        if (psav.null()) cerr << "Not Found!" << endl;
	else             cerr << "found!" << endl;
    }
    return psav;
}

//======================================  File primitives
int 
DaccIn::closeFile(void) {
    if (mDebug > 3) cout << "Closing frame file ...";
    if (mStream) {
        delete mStream;
	mStream = 0;
    }

    if (mSbuf) {
        delete mSbuf;
        mSbuf = 0;
	mFile.pop_front();
    }

    mSource = s_none;
    if (mDebug > 3) cout << " Done" << endl;
    return 0;
}

int 
DaccIn::openFile(const char* Source) {

    //----------------------------------  Request online data (ignore start)
    int nc = strlen(DACC_ONLDEV);
    if (!strncmp(Source, DACC_ONLDEV, nc)) {
#ifndef DMTOFFLINE
        if (mDebug > 3) cout << "Opening partition ..." << endl;
        iSMbuf* buf = new iSMbuf;
        mSbuf = buf->open(Source+nc, ios::in);
	if (!mSbuf) {
	    cerr << "Unable to open partition " << Source+nc << endl;
	    mFile.pop_front();
	    return -1;
	} else if (mDebug) {
	    cout << "Opened partition " << Source+nc << " for frame input." 
		 << endl;
	}
	mSource = s_online;
#else
	cerr << "Shared memory partitions not implemented in offline version"
	     << endl;
	return -1;
#endif

    //----------------------------------  Forget about offline for now
    } else {
        if (mDebug > 3) cout << "Opening file ..." << endl;
        filebuf* fbuf = new filebuf;
	mSbuf = fbuf->open((char*)Source, ios::in);
	if (!mSbuf) {
	    cerr << "Unable to open file " << Source << endl;
	    mFile.pop_front();
	    return -1;
	} else if (mDebug) {
	    cout << "Opened file " << Source << " for frame input." << endl;
	}
	mSource = s_file;
    }

    mStream = new std::istream(mSbuf);
    if (!mStream) return -1;
    return 0;
}

//======================================  Reader primitives
int 
DaccIn::closeReader(void) {
    int rc = 0;
    if (!mReader) return rc;
    mFramesInFile = 0;

    try {
        delete mReader;
    } catch (exception& e) {
        cerr << "Exception in DaccIn::close() when deleting IFrameStream: " 
	     << e.what() << endl;
	rc = -1;
    } catch (...) {
        cerr << "Unidentified exception caught." << endl;
	rc = -1;
    }
    mReader = 0;
    return rc;
}

int 
DaccIn::openReader(void) {
    if (mDebug > 3) cout << "Opening stream reader ...";
    try {
        mReader = new IFrameStream(*mStream);
    } catch (exception& e) {
        cerr << "Exception constructing IFrameStream: " << e.what()<< endl;
	return -1;
    } catch (...) {
        cerr << "Unidentified exception caught." << endl;
	return -1;
    }
    if (mDebug > 3) cout << " Done!" << endl;
    mFrameNo = 0;
    return 0;
}

//======================================  Frame primitives
int 
DaccIn::deleteFrame(void) {
    try {
        delete mFrame;
    } catch (exception& e) {
        cerr << "deleteFrame: Exception in frame destructor: " 
	     << e.what() << endl;
    } catch (...) {
        cerr << "deleteFrame: Unexpected exception." << endl;
    }
    mFrame = 0;
    mFrameNo++;
    return 0;
}

int 
DaccIn::readFrame(void) {
    if (mDebug > 3) cout << "Reading next frame...";
    if (haveFrame() || !isOpen()) return -1;
    try {
        FrRawData* raw;
        switch (mTOCMode) {
	case s_noTOC:
	    mFrame = mReader->ReadNextFrame();
	    if (!haveFrame()) {
	        if (!mFrameNo) cerr << "File has no frames!" << endl;
		return -1;
	    }
	    raw = mFrame->GetRawData();
	    if (raw) mAdcIter = raw->RefFirstAdc().begin();
	    break;
	case s_TOC:
	    if (mFrameNo && mFramesInFile <= mFrameNo) return -1;
	    unsigned int rflgs = FrameH::DETECT_SIM | FrameH::DETECT_PROC 
	                       | FrameH::HISTORY;
	    mFrame = mReader->ReadFrameH(mFrameNo, rflgs);
	    if (!haveFrame()) {
	        cerr << "Unable to read frame header!" << endl;
#ifdef ISMBUF_TRACE
		if (mSbuf && mSource == s_online) {
		    dynamic_cast<iSMbuf*>(mSbuf)->pTrace();
		}
#endif
		return -1;
	    }
	    raw = mReader->ReadFrRawData(mFrameNo);
	    if (raw) {
	        mFrame->SetRawData(raw, false, true);
		mAdcIter = raw->RefFirstAdc().begin();
	    }
	    if (!mFrameNo) mFramesInFile = mReader->GetNumberOfFrames();
	}
	if (mDebug > 2) cout << "Frame read successfully. ID: " 
			     <<  mFrame->GetFrame() << endl;
    } catch (exception& e) {
        mFrame = static_cast<FrameH*>(0);
	if (mDebug) {
	    cerr << "Exception reading Frame: " << e.what() << endl;
	    cerr << "Error occurred at offset " << mStream->tellg() << endl;
	}
	return -1;
    } catch (...) {
        mFrame = static_cast<FrameH*>(0);	
	if (mDebug) cerr << "Unidentified exception caught." << endl;
	return -1;
    }
    if (mDebug > 3) cout << " Done" << endl;
    return 0;
}

const FrAdcData*
DaccIn::readAdc(const string& name) const {
    FrRawData* raw = mFrame->GetRawData();
    if (!raw) {
        raw = new FrRawData();
	mFrame->SetRawData(raw, false, true);
    }
    if (getDebug() > 4) cout << "Slurp frame: " << mFrameNo 
			     << " Adc: " << name << endl;
    FrAdcData* p = 0;
    try {
	p = mReader->ReadFrAdcData(mFrameNo, name);
	if (p) mAdcIter = raw->RefFirstAdc().append(p, false, true);
	else   mAdcIter = raw->RefFirstAdc().end();
    } catch (exception& e) {
	p = 0;
	if (mDebug > 2) cerr << "Exception reading Adc " << name 
			     << ": " << e.what() << endl;
    }
    return p;
}

const FrProcData*
DaccIn::readProc(const string& name) const {
    if (getDebug() > 4) cout << "Slurp frame: " << mFrameNo 
			     << " FrProcData: " << name << endl;
    FrProcData* p = 0;
    try {
	p = mReader->ReadFrProcData(mFrameNo, name);
	if (p) mProcIter = mFrame->RefProcData().append(p, false, true);
	else   mProcIter = mFrame->RefProcData().end();
    } catch (exception& e) {
	p = 0;
	if (mDebug > 2) cerr << "Exception reading FrProcData " << name 
			     << ": " << e.what() << endl;
    }
    return p;
}

const FrSimData*
DaccIn::readSim(const string& name) const {
    if (mDebug > 4) cout << "Slurp frame: " << mFrameNo 
                             << " FrSimData: " << name << "... ";
    FrSimData* p = 0;
    try {
        p = mReader->ReadFrSimData(mFrameNo, name);
	if (mDebug > 4) {
	    if (p) cout << "Found!" << endl;
	    else   cout << "Not Found!" << endl;
	}
	if (p) mSimIter = mFrame->RefSimData().append(p, false, true);
  	else   mSimIter = mFrame->RefSimData().end();
    } catch (exception& e) {
        p = 0;
	if (mDebug > 4) cout << "Failed!" << endl;
	if (mDebug > 2) cerr << "Exception reading FrSimData " << name 
    			     << ": " << e.what() << endl;
    }
    return p;
}
